# test_config.py -- Tests for reading and writing configuration files
# Copyright (C) 2011 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for reading and writing configuration files."""

import os
import sys
import tempfile
from io import BytesIO
from unittest import skipIf
from unittest.mock import patch

from dulwich.config import (
    CaseInsensitiveOrderedMultiDict,
    ConfigDict,
    ConfigFile,
    StackedConfig,
    _check_section_name,
    _check_variable_name,
    _escape_value,
    _format_string,
    _parse_string,
    apply_instead_of,
    env_config,
    get_git_proxy_command,
    parse_submodules,
    read_submodules,
)

from . import TestCase


class ConfigFileTests(TestCase):
    def from_file(self, text):
        return ConfigFile.from_file(BytesIO(text))

    def test_empty(self) -> None:
        ConfigFile()

    def test_eq(self) -> None:
        self.assertEqual(ConfigFile(), ConfigFile())

    def test_default_config(self) -> None:
        cf = self.from_file(
            b"""[core]
\trepositoryformatversion = 0
\tfilemode = true
\tbare = false
\tlogallrefupdates = true
"""
        )
        self.assertEqual(
            ConfigFile(
                {
                    (b"core",): {
                        b"repositoryformatversion": b"0",
                        b"filemode": b"true",
                        b"bare": b"false",
                        b"logallrefupdates": b"true",
                    }
                }
            ),
            cf,
        )

    def test_from_file_empty(self) -> None:
        cf = self.from_file(b"")
        self.assertEqual(ConfigFile(), cf)

    def test_empty_line_before_section(self) -> None:
        cf = self.from_file(b"\n[section]\n")
        self.assertEqual(ConfigFile({(b"section",): {}}), cf)

    def test_comment_before_section(self) -> None:
        cf = self.from_file(b"# foo\n[section]\n")
        self.assertEqual(ConfigFile({(b"section",): {}}), cf)

    def test_comment_after_section(self) -> None:
        cf = self.from_file(b"[section] # foo\n")
        self.assertEqual(ConfigFile({(b"section",): {}}), cf)

    def test_comment_after_variable(self) -> None:
        cf = self.from_file(b"[section]\nbar= foo # a comment\n")
        self.assertEqual(ConfigFile({(b"section",): {b"bar": b"foo"}}), cf)

    def test_comment_character_within_value_string(self) -> None:
        cf = self.from_file(b'[section]\nbar= "foo#bar"\n')
        self.assertEqual(ConfigFile({(b"section",): {b"bar": b"foo#bar"}}), cf)

    def test_comment_character_within_section_string(self) -> None:
        cf = self.from_file(b'[branch "foo#bar"] # a comment\nbar= foo\n')
        self.assertEqual(ConfigFile({(b"branch", b"foo#bar"): {b"bar": b"foo"}}), cf)

    def test_closing_bracket_within_section_string(self) -> None:
        cf = self.from_file(b'[branch "foo]bar"] # a comment\nbar= foo\n')
        self.assertEqual(ConfigFile({(b"branch", b"foo]bar"): {b"bar": b"foo"}}), cf)

    def test_from_file_section(self) -> None:
        cf = self.from_file(b"[core]\nfoo = bar\n")
        self.assertEqual(b"bar", cf.get((b"core",), b"foo"))
        self.assertEqual(b"bar", cf.get((b"core", b"foo"), b"foo"))

    def test_from_file_multiple(self) -> None:
        cf = self.from_file(b"[core]\nfoo = bar\nfoo = blah\n")
        self.assertEqual([b"bar", b"blah"], list(cf.get_multivar((b"core",), b"foo")))
        self.assertEqual([], list(cf.get_multivar((b"core",), b"blah")))

    def test_from_file_utf8_bom(self) -> None:
        text = "[core]\nfoo = b\u00e4r\n".encode("utf-8-sig")
        cf = self.from_file(text)
        self.assertEqual(b"b\xc3\xa4r", cf.get((b"core",), b"foo"))

    def test_from_file_section_case_insensitive_lower(self) -> None:
        cf = self.from_file(b"[cOre]\nfOo = bar\n")
        self.assertEqual(b"bar", cf.get((b"core",), b"foo"))
        self.assertEqual(b"bar", cf.get((b"core", b"foo"), b"foo"))

    def test_from_file_section_case_insensitive_mixed(self) -> None:
        cf = self.from_file(b"[cOre]\nfOo = bar\n")
        self.assertEqual(b"bar", cf.get((b"core",), b"fOo"))
        self.assertEqual(b"bar", cf.get((b"cOre", b"fOo"), b"fOo"))

    def test_from_file_with_mixed_quoted(self) -> None:
        cf = self.from_file(b'[core]\nfoo = "bar"la\n')
        self.assertEqual(b"barla", cf.get((b"core",), b"foo"))

    def test_from_file_section_with_open_brackets(self) -> None:
        self.assertRaises(ValueError, self.from_file, b"[core\nfoo = bar\n")

    def test_from_file_value_with_open_quoted(self) -> None:
        self.assertRaises(ValueError, self.from_file, b'[core]\nfoo = "bar\n')

    def test_from_file_with_quotes(self) -> None:
        cf = self.from_file(b'[core]\nfoo = " bar"\n')
        self.assertEqual(b" bar", cf.get((b"core",), b"foo"))

    def test_from_file_with_interrupted_line(self) -> None:
        cf = self.from_file(b"[core]\nfoo = bar\\\n la\n")
        self.assertEqual(b"barla", cf.get((b"core",), b"foo"))

    def test_from_file_with_boolean_setting(self) -> None:
        cf = self.from_file(b"[core]\nfoo\n")
        self.assertEqual(b"true", cf.get((b"core",), b"foo"))

    def test_from_file_subsection(self) -> None:
        cf = self.from_file(b'[branch "foo"]\nfoo = bar\n')
        self.assertEqual(b"bar", cf.get((b"branch", b"foo"), b"foo"))

    def test_from_file_subsection_invalid(self) -> None:
        self.assertRaises(ValueError, self.from_file, b'[branch "foo]\nfoo = bar\n')

    def test_from_file_subsection_not_quoted(self) -> None:
        cf = self.from_file(b"[branch.foo]\nfoo = bar\n")
        self.assertEqual(b"bar", cf.get((b"branch", b"foo"), b"foo"))

    def test_from_file_includeif_hasconfig(self) -> None:
        """Test parsing includeIf sections with hasconfig conditions."""
        # Test case from issue #1216
        cf = self.from_file(
            b'[includeIf "hasconfig:remote.*.url:ssh://org-*@github.com/**"]\n'
            b"    path = ~/.config/git/.work\n"
        )
        self.assertEqual(
            b"~/.config/git/.work",
            cf.get(
                (b"includeIf", b"hasconfig:remote.*.url:ssh://org-*@github.com/**"),
                b"path",
            ),
        )

    def test_write_preserve_multivar(self) -> None:
        cf = self.from_file(b"[core]\nfoo = bar\nfoo = blah\n")
        f = BytesIO()
        cf.write_to_file(f)
        self.assertEqual(b"[core]\n\tfoo = bar\n\tfoo = blah\n", f.getvalue())

    def test_write_to_file_empty(self) -> None:
        c = ConfigFile()
        f = BytesIO()
        c.write_to_file(f)
        self.assertEqual(b"", f.getvalue())

    def test_write_to_file_section(self) -> None:
        c = ConfigFile()
        c.set((b"core",), b"foo", b"bar")
        f = BytesIO()
        c.write_to_file(f)
        self.assertEqual(b"[core]\n\tfoo = bar\n", f.getvalue())

    def test_write_to_file_section_multiple(self) -> None:
        c = ConfigFile()
        c.set((b"core",), b"foo", b"old")
        c.set((b"core",), b"foo", b"new")
        f = BytesIO()
        c.write_to_file(f)
        self.assertEqual(b"[core]\n\tfoo = new\n", f.getvalue())

    def test_write_to_file_subsection(self) -> None:
        c = ConfigFile()
        c.set((b"branch", b"blie"), b"foo", b"bar")
        f = BytesIO()
        c.write_to_file(f)
        self.assertEqual(b'[branch "blie"]\n\tfoo = bar\n', f.getvalue())

    def test_write_to_file_subsection_escapes_quote_and_backslash(self) -> None:
        c = ConfigFile()
        c.set((b"branch", b'foo"bar'), b"x", b"1")
        c.set((b"branch", b"a\\b"), b"y", b"2")
        f = BytesIO()
        c.write_to_file(f)
        self.assertEqual(
            b'[branch "foo\\"bar"]\n\tx = 1\n[branch "a\\\\b"]\n\ty = 2\n',
            f.getvalue(),
        )

    def test_subsection_roundtrip_with_special_chars(self) -> None:
        for name in (b'foo"bar', b"a\\b", b'a\\"b', b"trailing\\"):
            c = ConfigFile()
            c.set((b"branch", name), b"k", b"v")
            f = BytesIO()
            c.write_to_file(f)
            reparsed = self.from_file(f.getvalue())
            self.assertEqual(b"v", reparsed.get((b"branch", name), b"k"))

    def test_write_to_file_subsection_rejects_newline(self) -> None:
        c = ConfigFile()
        c.set((b"branch", b"foo\nbar"), b"k", b"v")
        self.assertRaises(ValueError, c.write_to_file, BytesIO())

    def test_from_file_subsection_unknown_escape_is_lenient(self) -> None:
        # Git silently drops the backslash on unrecognised escape sequences
        # in subsection names; we match that behaviour for compatibility
        # with includeIf headers containing literal Windows backslashes.
        cf = self.from_file(b'[branch "foo\\nbar"]\nk = v\n')
        self.assertEqual(b"v", cf.get((b"branch", b"foonbar"), b"k"))

    def test_write_to_file_preserves_quoted_trailing_whitespace(self) -> None:
        c = ConfigFile()
        c.set((b"core",), b"foo", b" ")
        c.set((b"core",), b"bar", b"\t")

        f = BytesIO()
        c.write_to_file(f)
        reparsed = self.from_file(f.getvalue())

        self.assertEqual(c, reparsed)

    def test_same_line(self) -> None:
        cf = self.from_file(b"[branch.foo] foo = bar\n")
        self.assertEqual(b"bar", cf.get((b"branch", b"foo"), b"foo"))

    def test_quoted_newlines_windows(self) -> None:
        cf = self.from_file(
            b"[alias]\r\n"
            b"c = '!f() { \\\r\n"
            b' printf \'[git commit -m \\"%s\\"]\\n\' \\"$*\\" && \\\r\n'
            b' git commit -m \\"$*\\"; \\\r\n'
            b" }; f'\r\n"
        )
        self.assertEqual(list(cf.sections()), [(b"alias",)])
        self.assertEqual(
            b'\'!f() { printf \'[git commit -m "%s"]\n\' "$*" && git commit -m "$*"',
            cf.get((b"alias",), b"c"),
        )

    def test_quoted(self) -> None:
        cf = self.from_file(
            b"""[gui]
\tfontdiff = -family \\\"Ubuntu Mono\\\" -size 11 -overstrike 0
"""
        )
        self.assertEqual(
            ConfigFile(
                {
                    (b"gui",): {
                        b"fontdiff": b'-family "Ubuntu Mono" -size 11 -overstrike 0',
                    }
                }
            ),
            cf,
        )

    def test_quoted_multiline(self) -> None:
        cf = self.from_file(
            b"""[alias]
who = \"!who() {\\
  git log --no-merges --pretty=format:'%an - %ae' $@ | uniq -c | sort -rn;\\
};\\
who\"
"""
        )
        self.assertEqual(
            ConfigFile(
                {
                    (b"alias",): {
                        b"who": (
                            b"!who() {git log --no-merges --pretty=format:'%an - "
                            b"%ae' $@ | uniq -c | sort -rn;};who"
                        )
                    }
                }
            ),
            cf,
        )

    def test_set_hash_gets_quoted(self) -> None:
        c = ConfigFile()
        c.set(b"xandikos", b"color", b"#665544")
        f = BytesIO()
        c.write_to_file(f)
        self.assertEqual(b'[xandikos]\n\tcolor = "#665544"\n', f.getvalue())

    def test_windows_path_with_trailing_backslash_unquoted(self) -> None:
        """Test that Windows paths ending with escaped backslash are handled correctly."""
        # This reproduces the issue from https://github.com/jelmer/dulwich/issues/1088
        # A single backslash at the end should actually be a line continuation in strict Git config
        # But we want to be more tolerant like Git itself
        cf = self.from_file(
            b'[core]\n\trepositoryformatversion = 0\n[remote "origin"]\n\turl = C:/Users/test\\\\\n\tfetch = +refs/heads/*:refs/remotes/origin/*\n'
        )
        self.assertEqual(b"C:/Users/test\\", cf.get((b"remote", b"origin"), b"url"))
        self.assertEqual(
            b"+refs/heads/*:refs/remotes/origin/*",
            cf.get((b"remote", b"origin"), b"fetch"),
        )

    def test_windows_path_with_trailing_backslash_quoted(self) -> None:
        """Test that quoted Windows paths with escaped backslashes work correctly."""
        cf = self.from_file(
            b'[core]\n\trepositoryformatversion = 0\n[remote "origin"]\n\turl = "C:\\\\Users\\\\test\\\\"\n\tfetch = +refs/heads/*:refs/remotes/origin/*\n'
        )
        self.assertEqual(b"C:\\Users\\test\\", cf.get((b"remote", b"origin"), b"url"))
        self.assertEqual(
            b"+refs/heads/*:refs/remotes/origin/*",
            cf.get((b"remote", b"origin"), b"fetch"),
        )

    def test_single_backslash_at_line_end_shows_proper_escaping_needed(self) -> None:
        """Test that demonstrates proper escaping is needed for single backslashes."""
        # This test documents the current behavior: a single backslash at the end of a line
        # is treated as a line continuation per Git config spec. Users should escape backslashes.

        # This reproduces the original issue - single backslash causes line continuation
        cf = self.from_file(
            b'[remote "origin"]\n\turl = C:/Users/test\\\n\tfetch = +refs/heads/*:refs/remotes/origin/*\n'
        )
        # The result shows that line continuation occurred
        self.assertEqual(
            b"C:/Users/testfetch = +refs/heads/*:refs/remotes/origin/*",
            cf.get((b"remote", b"origin"), b"url"),
        )

        # The proper way to include a literal backslash is to escape it
        cf2 = self.from_file(
            b'[remote "origin"]\n\turl = C:/Users/test\\\\\n\tfetch = +refs/heads/*:refs/remotes/origin/*\n'
        )
        self.assertEqual(b"C:/Users/test\\", cf2.get((b"remote", b"origin"), b"url"))
        self.assertEqual(
            b"+refs/heads/*:refs/remotes/origin/*",
            cf2.get((b"remote", b"origin"), b"fetch"),
        )

    def test_from_path_pathlib(self) -> None:
        import tempfile
        from pathlib import Path

        # Create a temporary config file
        with tempfile.NamedTemporaryFile(mode="w", suffix=".config", delete=False) as f:
            f.write("[core]\n    filemode = true\n")
            temp_path = f.name

        self.addCleanup(os.unlink, temp_path)

        # Test with pathlib.Path
        path_obj = Path(temp_path)
        cf = ConfigFile.from_path(path_obj)
        self.assertEqual(cf.get((b"core",), b"filemode"), b"true")

    def test_write_to_path_pathlib(self) -> None:
        import tempfile
        from pathlib import Path

        # Create a config
        cf = ConfigFile()
        cf.set((b"user",), b"name", b"Test User")

        # Write to pathlib.Path
        with tempfile.NamedTemporaryFile(suffix=".config", delete=False) as f:
            temp_path = f.name

        try:
            path_obj = Path(temp_path)
            cf.write_to_path(path_obj)

            # Read it back
            cf2 = ConfigFile.from_path(path_obj)
            self.assertEqual(cf2.get((b"user",), b"name"), b"Test User")
        finally:
            # Clean up
            os.unlink(temp_path)

    def test_include_basic(self) -> None:
        """Test basic include functionality."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create included config file
            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(
                    b"[user]\n    name = Included User\n    email = included@example.com\n"
                )

            # Create main config with include
            main_config = self.from_file(
                b"[user]\n    name = Main User\n[include]\n    path = included.config\n"
            )

            # Should not include anything without proper directory context
            self.assertEqual(b"Main User", main_config.get((b"user",), b"name"))
            with self.assertRaises(KeyError):
                main_config.get((b"user",), b"email")

            # Now test with proper file loading
            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                f.write(
                    b"[user]\n    name = Main User\n[include]\n    path = included.config\n"
                )

            # Load from path to get include functionality
            cf = ConfigFile.from_path(main_path)
            self.assertEqual(b"Included User", cf.get((b"user",), b"name"))
            self.assertEqual(b"included@example.com", cf.get((b"user",), b"email"))

    def test_include_absolute_path(self) -> None:
        """Test include with absolute path."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Use realpath to resolve any symlinks (important on macOS and Windows)
            tmpdir = os.path.realpath(tmpdir)

            # Create included config file
            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(b"[core]\n    bare = true\n")

            # Create main config with absolute include path
            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                # Properly escape backslashes in Windows paths
                escaped_path = included_path.replace("\\", "\\\\")
                f.write(f"[include]\n    path = {escaped_path}\n".encode())

            cf = ConfigFile.from_path(main_path)
            self.assertEqual(b"true", cf.get((b"core",), b"bare"))

    def test_expand_includes_disabled(self) -> None:
        """include directives are ignored when expand_includes=False."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = os.path.realpath(tmpdir)

            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(b"[core]\n    bare = true\n")

            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                escaped_path = included_path.replace("\\", "\\\\")
                f.write(f"[include]\n    path = {escaped_path}\n".encode())

            cf = ConfigFile.from_path(main_path, expand_includes=False)
            self.assertRaises(KeyError, cf.get, (b"core",), b"bare")

    def test_includeif_hasconfig(self) -> None:
        """Test includeIf with hasconfig conditions."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create included config file
            work_included_path = os.path.join(tmpdir, "work.config")
            with open(work_included_path, "wb") as f:
                f.write(b"[user]\n    email = work@company.com\n")

            personal_included_path = os.path.join(tmpdir, "personal.config")
            with open(personal_included_path, "wb") as f:
                f.write(b"[user]\n    email = personal@example.com\n")

            # Create main config with hasconfig conditions
            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                f.write(
                    b'[remote "origin"]\n'
                    b"    url = ssh://org-work@github.com/company/project\n"
                    b'[includeIf "hasconfig:remote.*.url:ssh://org-*@github.com/**"]\n'
                    b"    path = work.config\n"
                    b'[includeIf "hasconfig:remote.*.url:https://github.com/opensource/**"]\n'
                    b"    path = personal.config\n"
                )

            # Load config - should match the work config due to org-work remote
            # The second condition won't match since url doesn't have /opensource/ path
            cf = ConfigFile.from_path(main_path)
            self.assertEqual(b"work@company.com", cf.get((b"user",), b"email"))

    def test_includeif_hasconfig_wildcard(self) -> None:
        """Test includeIf hasconfig with wildcard patterns."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create included config
            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(b"[user]\n    name = IncludedUser\n")

            # Create main config with hasconfig condition using wildcards
            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                f.write(
                    b"[core]\n"
                    b"    autocrlf = true\n"
                    b'[includeIf "hasconfig:core.autocrlf:true"]\n'
                    b"    path = included.config\n"
                )

            # Load config - should include based on core.autocrlf value
            cf = ConfigFile.from_path(main_path)
            self.assertEqual(b"IncludedUser", cf.get((b"user",), b"name"))

    def test_includeif_hasconfig_no_match(self) -> None:
        """Test includeIf hasconfig when condition doesn't match."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create included config
            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(b"[user]\n    name = IncludedUser\n")

            # Create main config with non-matching hasconfig condition
            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                f.write(
                    b"[core]\n"
                    b"    autocrlf = false\n"
                    b'[includeIf "hasconfig:core.autocrlf:true"]\n'
                    b"    path = included.config\n"
                )

            # Load config - should NOT include since condition doesn't match
            cf = ConfigFile.from_path(main_path)
            with self.assertRaises(KeyError):
                cf.get((b"user",), b"name")

    def test_include_circular(self) -> None:
        """Test that circular includes are handled properly."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create two configs that include each other
            config1_path = os.path.join(tmpdir, "config1")
            config2_path = os.path.join(tmpdir, "config2")

            with open(config1_path, "wb") as f:
                f.write(b"[user]\n    name = User1\n[include]\n    path = config2\n")

            with open(config2_path, "wb") as f:
                f.write(
                    b"[user]\n    email = user2@example.com\n[include]\n    path = config1\n"
                )

            # Should handle circular includes gracefully
            cf = ConfigFile.from_path(config1_path)
            self.assertEqual(b"User1", cf.get((b"user",), b"name"))
            self.assertEqual(b"user2@example.com", cf.get((b"user",), b"email"))

    def test_include_missing_file(self) -> None:
        """Test that missing include files are ignored."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create config with include of non-existent file
            config_path = os.path.join(tmpdir, "config")
            with open(config_path, "wb") as f:
                f.write(
                    b"[user]\n    name = TestUser\n[include]\n    path = missing.config\n"
                )

            # Should not fail, just ignore missing include
            cf = ConfigFile.from_path(config_path)
            self.assertEqual(b"TestUser", cf.get((b"user",), b"name"))

    def test_include_depth_limit(self) -> None:
        """Test that excessive include depth is prevented."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create a chain of includes that exceeds depth limit
            for i in range(15):
                config_path = os.path.join(tmpdir, f"config{i}")
                with open(config_path, "wb") as f:
                    if i == 0:
                        f.write(b"[user]\n    name = User0\n")
                    f.write(f"[include]\n    path = config{i + 1}\n".encode())

            # Should raise error due to depth limit
            with self.assertRaises(ValueError) as cm:
                ConfigFile.from_path(os.path.join(tmpdir, "config0"))
            self.assertIn("include depth", str(cm.exception))

    def test_include_with_custom_file_opener(self) -> None:
        """Test include functionality with a custom file opener for security."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create config files
            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(b"[user]\n    email = custom@example.com\n")

            restricted_path = os.path.join(tmpdir, "restricted.config")
            with open(restricted_path, "wb") as f:
                f.write(b"[user]\n    email = restricted@example.com\n")

            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                f.write(b"[user]\n    name = Test User\n")
                f.write(b"[include]\n    path = included.config\n")
                f.write(b"[include]\n    path = restricted.config\n")

            # Define a custom file opener that restricts access
            allowed_files = {included_path, main_path}

            def secure_file_opener(path):
                path_str = os.fspath(path)
                if path_str not in allowed_files:
                    raise PermissionError(f"Access denied to {path}")
                return open(path_str, "rb")

            # Load config with restricted file access
            cf = ConfigFile.from_path(main_path, file_opener=secure_file_opener)

            # Should have the main config and included config, but not restricted
            self.assertEqual(b"Test User", cf.get((b"user",), b"name"))
            self.assertEqual(b"custom@example.com", cf.get((b"user",), b"email"))
            # Email from restricted.config should not be loaded

    def test_unknown_includeif_condition(self) -> None:
        """Test that unknown includeIf conditions are silently ignored (like Git)."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create included config file
            included_path = os.path.join(tmpdir, "included.config")
            with open(included_path, "wb") as f:
                f.write(b"[user]\n    email = included@example.com\n")

            # Create main config with unknown includeIf condition
            main_path = os.path.join(tmpdir, "main.config")
            with open(main_path, "wb") as f:
                f.write(b"[user]\n    name = Main User\n")
                f.write(
                    b'[includeIf "unknowncondition:foo"]\n    path = included.config\n'
                )

            # Should not fail, just ignore the unknown condition
            cf = ConfigFile.from_path(main_path)
            self.assertEqual(b"Main User", cf.get((b"user",), b"name"))
            # Email should not be included because condition is unknown
            with self.assertRaises(KeyError):
                cf.get((b"user",), b"email")

    def test_missing_include_file_logging(self) -> None:
        """Test that missing include files are logged but don't cause failure."""
        import logging
        from io import StringIO

        # Set up logging capture
        log_capture = StringIO()
        handler = logging.StreamHandler(log_capture)
        handler.setLevel(logging.DEBUG)
        logger = logging.getLogger("dulwich.config")
        old_level = logger.level
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                config_path = os.path.join(tmpdir, "test.config")
                with open(config_path, "wb") as f:
                    f.write(b"[user]\n    name = Test User\n")
                    f.write(b"[include]\n    path = nonexistent.config\n")

                # Should not fail, just log
                cf = ConfigFile.from_path(config_path)
                self.assertEqual(b"Test User", cf.get((b"user",), b"name"))

                # Check that it was logged
                log_output = log_capture.getvalue()
                self.assertIn("Invalid include path", log_output)
                self.assertIn("nonexistent.config", log_output)
        finally:
            logger.removeHandler(handler)
            logger.setLevel(old_level)

    def test_invalid_include_path_logging(self) -> None:
        """Test that invalid include paths are logged but don't cause failure."""
        import logging
        from io import StringIO

        # Set up logging capture
        log_capture = StringIO()
        handler = logging.StreamHandler(log_capture)
        handler.setLevel(logging.DEBUG)
        logger = logging.getLogger("dulwich.config")
        old_level = logger.level
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                config_path = os.path.join(tmpdir, "test.config")
                with open(config_path, "wb") as f:
                    f.write(b"[user]\n    name = Test User\n")
                    # Use null bytes which are invalid in paths
                    f.write(b"[include]\n    path = /invalid\x00path/file.config\n")

                # Should not fail, just log
                cf = ConfigFile.from_path(config_path)
                self.assertEqual(b"Test User", cf.get((b"user",), b"name"))

                # Check that it was logged
                log_output = log_capture.getvalue()
                self.assertIn("Invalid include path", log_output)
        finally:
            logger.removeHandler(handler)
            logger.setLevel(old_level)

    def test_unknown_includeif_condition_logging(self) -> None:
        """Test that unknown includeIf conditions are logged."""
        import logging
        from io import StringIO

        # Set up logging capture
        log_capture = StringIO()
        handler = logging.StreamHandler(log_capture)
        handler.setLevel(logging.DEBUG)
        logger = logging.getLogger("dulwich.config")
        old_level = logger.level
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                config_path = os.path.join(tmpdir, "test.config")
                with open(config_path, "wb") as f:
                    f.write(b"[user]\n    name = Test User\n")
                    f.write(
                        b'[includeIf "futurefeature:value"]\n    path = other.config\n'
                    )

                # Should not fail, just log
                cf = ConfigFile.from_path(config_path)
                self.assertEqual(b"Test User", cf.get((b"user",), b"name"))

                # Check that it was logged
                log_output = log_capture.getvalue()
                self.assertIn("Unknown includeIf condition", log_output)
                self.assertIn("futurefeature:value", log_output)
        finally:
            logger.removeHandler(handler)
            logger.setLevel(old_level)

    def test_custom_file_opener_with_include_depth(self) -> None:
        """Test that custom file opener is passed through include chain."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Use realpath to resolve any symlinks
            tmpdir = os.path.realpath(tmpdir)

            # Create a chain of includes
            final_config = os.path.join(tmpdir, "final.config")
            with open(final_config, "wb") as f:
                f.write(b"[feature]\n    enabled = true\n")

            middle_config = os.path.join(tmpdir, "middle.config")
            with open(middle_config, "wb") as f:
                f.write(b"[user]\n    email = test@example.com\n")
                escaped_final_config = final_config.replace("\\", "\\\\")
                f.write(f"[include]\n    path = {escaped_final_config}\n".encode())

            main_config = os.path.join(tmpdir, "main.config")
            with open(main_config, "wb") as f:
                f.write(b"[user]\n    name = Test User\n")
                escaped_middle_config = middle_config.replace("\\", "\\\\")
                f.write(f"[include]\n    path = {escaped_middle_config}\n".encode())

            # Track file access order
            access_order = []

            def ordering_file_opener(path):
                path_str = os.fspath(path)
                access_order.append(os.path.basename(path_str))
                return open(path_str, "rb")

            # Load config
            cf = ConfigFile.from_path(main_config, file_opener=ordering_file_opener)

            # Verify all values were loaded
            self.assertEqual(b"Test User", cf.get((b"user",), b"name"))
            self.assertEqual(b"test@example.com", cf.get((b"user",), b"email"))
            self.assertEqual(b"true", cf.get((b"feature",), b"enabled"))

            # Verify access order
            self.assertEqual(
                ["main.config", "middle.config", "final.config"], access_order
            )


class ConfigDictTests(TestCase):
    def test_get_set(self) -> None:
        cd = ConfigDict()
        self.assertRaises(KeyError, cd.get, b"foo", b"core")
        cd.set((b"core",), b"foo", b"bla")
        self.assertEqual(b"bla", cd.get((b"core",), b"foo"))
        cd.set((b"core",), b"foo", b"bloe")
        self.assertEqual(b"bloe", cd.get((b"core",), b"foo"))

    def test_get_boolean(self) -> None:
        cd = ConfigDict()
        cd.set((b"core",), b"foo", b"true")
        self.assertTrue(cd.get_boolean((b"core",), b"foo"))
        cd.set((b"core",), b"foo", b"false")
        self.assertFalse(cd.get_boolean((b"core",), b"foo"))
        cd.set((b"core",), b"foo", b"invalid")
        self.assertRaises(ValueError, cd.get_boolean, (b"core",), b"foo")

    def test_dict(self) -> None:
        cd = ConfigDict()
        cd.set((b"core",), b"foo", b"bla")
        cd.set((b"core2",), b"foo", b"bloe")

        self.assertEqual([(b"core",), (b"core2",)], list(cd.keys()))
        self.assertEqual(cd[(b"core",)], {b"foo": b"bla"})

        cd[b"a"] = b"b"
        self.assertEqual(cd[b"a"], b"b")

    def test_items(self) -> None:
        cd = ConfigDict()
        cd.set((b"core",), b"foo", b"bla")
        cd.set((b"core2",), b"foo", b"bloe")

        self.assertEqual([(b"foo", b"bla")], list(cd.items((b"core",))))

    def test_items_nonexistant(self) -> None:
        cd = ConfigDict()
        cd.set((b"core2",), b"foo", b"bloe")

        self.assertEqual([], list(cd.items((b"core",))))

    def test_sections(self) -> None:
        cd = ConfigDict()
        cd.set((b"core2",), b"foo", b"bloe")

        self.assertEqual([(b"core2",)], list(cd.sections()))

    def test_set_vs_add(self) -> None:
        cd = ConfigDict()
        # Test add() creates multivars
        cd.add((b"core",), b"foo", b"value1")
        cd.add((b"core",), b"foo", b"value2")
        self.assertEqual(
            [b"value1", b"value2"], list(cd.get_multivar((b"core",), b"foo"))
        )

        # Test set() replaces values
        cd.set((b"core",), b"foo", b"value3")
        self.assertEqual([b"value3"], list(cd.get_multivar((b"core",), b"foo")))
        self.assertEqual(b"value3", cd.get((b"core",), b"foo"))


class StackedConfigTests(TestCase):
    def test_default_backends(self) -> None:
        StackedConfig.default_backends()

    @skipIf(sys.platform != "win32", "Windows specific config location.")
    def test_windows_config_from_path(self) -> None:
        from dulwich.config import get_win_system_paths

        install_dir = os.path.join("C:", "foo", "Git")
        self.overrideEnv("PATH", os.path.join(install_dir, "cmd"))
        with patch("os.path.exists", return_value=True):
            paths = list(get_win_system_paths())
        # Should only return the current Git config path, not legacy paths
        self.assertEqual(
            [os.path.join(install_dir, "etc", "gitconfig")],
            paths,
        )

    @skipIf(sys.platform != "win32", "Windows specific config location.")
    def test_windows_config_from_reg(self) -> None:
        import winreg

        from dulwich.config import get_win_system_paths

        self.overrideEnv("PATH", None)
        install_dir = os.path.join("C:", "foo", "Git")
        with patch("winreg.OpenKey"):
            with patch(
                "winreg.QueryValueEx",
                return_value=(install_dir, winreg.REG_SZ),
            ):
                paths = list(get_win_system_paths())
        # Should only return the current Git config path, not legacy paths
        self.assertEqual(
            [os.path.join(install_dir, "etc", "gitconfig")],
            paths,
        )

    @skipIf(sys.platform != "win32", "Windows specific config location.")
    def test_windows_legacy_paths(self) -> None:
        from dulwich.config import get_win_legacy_system_paths

        install_dir = os.path.join("C:", "foo", "Git")
        self.overrideEnv("PATH", os.path.join(install_dir, "cmd"))
        self.overrideEnv("PROGRAMDATA", "C:\\ProgramData")
        with patch("os.path.exists", return_value=True):
            paths = list(get_win_legacy_system_paths())
        # Should return all paths including legacy PROGRAMDATA location
        self.assertIn(
            os.path.join("C:\\ProgramData", "Git", "config"),
            paths,
        )
        self.assertIn(
            os.path.join(install_dir, "etc", "gitconfig"),
            paths,
        )

    def test_git_config_global(self) -> None:
        with tempfile.NamedTemporaryFile(
            suffix=".gitconfig", delete=False
        ) as global_config:
            # Write test config
            global_config.write(b"[user]\n\tname = Test User\n")
            global_config.flush()
            config_path = global_config.name

        try:
            self.overrideEnv("GIT_CONFIG_GLOBAL", config_path)
            backends = StackedConfig.default_backends()

            # Should have at least one backend (the global config)
            self.assertGreater(len(backends), 0)

            # Check that the custom global config is loaded
            paths = [backend.path for backend in backends if hasattr(backend, "path")]
            self.assertIn(
                config_path.encode() if isinstance(paths[0], bytes) else config_path,
                paths,
            )
        finally:
            os.unlink(config_path)

    def test_git_config_system(self) -> None:
        with tempfile.NamedTemporaryFile(
            suffix=".gitconfig", delete=False
        ) as system_config:
            # Write test config
            system_config.write(b"[core]\n\tautocrlf = true\n")
            system_config.flush()
            config_path = system_config.name

        try:
            self.overrideEnv("GIT_CONFIG_SYSTEM", config_path)
            self.overrideEnv("GIT_CONFIG_NOSYSTEM", None)  # Ensure it's not set
            backends = StackedConfig.default_backends()

            # Should have at least one backend (the system config)
            self.assertGreater(len(backends), 0)

            # Check that the custom system config is loaded
            paths = [backend.path for backend in backends if hasattr(backend, "path")]
            self.assertIn(
                config_path.encode() if isinstance(paths[0], bytes) else config_path,
                paths,
            )
        finally:
            os.unlink(config_path)

    def test_debug_logging(self) -> None:
        """Test that debug logging outputs gitconfig paths."""
        import logging

        # Set up a custom handler to capture log messages
        log_messages = []
        handler = logging.Handler()
        handler.emit = lambda record: log_messages.append(record.getMessage())

        # Get the dulwich.config logger and add our handler
        logger = logging.getLogger("dulwich.config")
        old_level = logger.level
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

        try:
            # Call default_backends which now logs unconditionally
            StackedConfig.default_backends()

            # Check that debug messages were logged
            self.assertTrue(
                any("Loading gitconfig from paths:" in msg for msg in log_messages),
                "Debug logging should log gitconfig paths",
            )
        finally:
            # Clean up
            logger.removeHandler(handler)
            logger.setLevel(old_level)

    def test_git_config_nosystem_precedence(self) -> None:
        # GIT_CONFIG_SYSTEM should take precedence over GIT_CONFIG_NOSYSTEM
        with tempfile.NamedTemporaryFile(
            suffix=".gitconfig", delete=False
        ) as system_config:
            system_config.write(b"[core]\n\ttest = value\n")
            system_config.flush()
            config_path = system_config.name

        try:
            self.overrideEnv("GIT_CONFIG_SYSTEM", config_path)
            self.overrideEnv("GIT_CONFIG_NOSYSTEM", "1")  # This should be ignored
            backends = StackedConfig.default_backends()

            # Check that the custom system config is still loaded
            paths = [backend.path for backend in backends if hasattr(backend, "path")]
            self.assertIn(
                config_path.encode() if isinstance(paths[0], bytes) else config_path,
                paths,
            )
        finally:
            os.unlink(config_path)


class EnvConfigTests(TestCase):
    def test_unset(self) -> None:
        self.assertIsNone(env_config({}))

    def test_empty_count(self) -> None:
        self.assertIsNone(
            env_config({"GIT_CONFIG_COUNT": ""}),
        )

    def test_zero(self) -> None:
        cf = env_config({"GIT_CONFIG_COUNT": "0"})
        self.assertIsNotNone(cf)
        self.assertEqual([], list(cf.sections()))

    def test_simple_key(self) -> None:
        cf = env_config(
            {
                "GIT_CONFIG_COUNT": "1",
                "GIT_CONFIG_KEY_0": "user.name",
                "GIT_CONFIG_VALUE_0": "Alice",
            }
        )
        self.assertEqual(b"Alice", cf.get((b"user",), b"name"))

    def test_subsection_key(self) -> None:
        cf = env_config(
            {
                "GIT_CONFIG_COUNT": "1",
                "GIT_CONFIG_KEY_0": "remote.upstream.url",
                "GIT_CONFIG_VALUE_0": "https://example.com/repo.git",
            }
        )
        self.assertEqual(
            b"https://example.com/repo.git",
            cf.get((b"remote", b"upstream"), b"url"),
        )

    def test_subsection_with_dots(self) -> None:
        # Subsection names may contain dots (e.g. url.<base>.insteadOf).
        cf = env_config(
            {
                "GIT_CONFIG_COUNT": "1",
                "GIT_CONFIG_KEY_0": "url.https://example.com/.insteadOf",
                "GIT_CONFIG_VALUE_0": "git@example.com:",
            }
        )
        self.assertEqual(
            b"git@example.com:",
            cf.get((b"url", b"https://example.com/"), b"insteadOf"),
        )

    def test_multi_value(self) -> None:
        cf = env_config(
            {
                "GIT_CONFIG_COUNT": "2",
                "GIT_CONFIG_KEY_0": "remote.origin.fetch",
                "GIT_CONFIG_VALUE_0": "+refs/heads/*:refs/remotes/origin/*",
                "GIT_CONFIG_KEY_1": "remote.origin.fetch",
                "GIT_CONFIG_VALUE_1": "+refs/tags/*:refs/tags/*",
            }
        )
        self.assertEqual(
            [
                b"+refs/heads/*:refs/remotes/origin/*",
                b"+refs/tags/*:refs/tags/*",
            ],
            list(cf.get_multivar((b"remote", b"origin"), b"fetch")),
        )

    def test_bogus_count_non_numeric(self) -> None:
        self.assertRaises(
            ValueError,
            env_config,
            {"GIT_CONFIG_COUNT": "abc"},
        )

    def test_bogus_count_negative(self) -> None:
        self.assertRaises(
            ValueError,
            env_config,
            {"GIT_CONFIG_COUNT": "-1"},
        )

    def test_missing_key(self) -> None:
        self.assertRaises(
            KeyError,
            env_config,
            {"GIT_CONFIG_COUNT": "1", "GIT_CONFIG_VALUE_0": "x"},
        )

    def test_missing_value(self) -> None:
        self.assertRaises(
            KeyError,
            env_config,
            {"GIT_CONFIG_COUNT": "1", "GIT_CONFIG_KEY_0": "user.name"},
        )

    def test_invalid_key_no_dot(self) -> None:
        self.assertRaises(
            ValueError,
            env_config,
            {
                "GIT_CONFIG_COUNT": "1",
                "GIT_CONFIG_KEY_0": "nodot",
                "GIT_CONFIG_VALUE_0": "x",
            },
        )

    def test_default_does_not_consult_env_count(self) -> None:
        # StackedConfig.default() must not silently pick up GIT_CONFIG_COUNT
        # overrides; callers have to opt in via env_overrides().
        with tempfile.NamedTemporaryFile(
            suffix=".gitconfig", delete=False
        ) as global_config:
            global_config.write(b"[user]\n\tname = FromFile\n")
            global_config.flush()
            config_path = global_config.name

        try:
            self.overrideEnv("GIT_CONFIG_GLOBAL", config_path)
            self.overrideEnv("GIT_CONFIG_NOSYSTEM", "1")
            self.overrideEnv("GIT_CONFIG_COUNT", "1")
            self.overrideEnv("GIT_CONFIG_KEY_0", "user.name")
            self.overrideEnv("GIT_CONFIG_VALUE_0", "FromEnv")
            sc = StackedConfig.default()
            self.assertEqual(b"FromFile", sc.get((b"user",), b"name"))
        finally:
            os.unlink(config_path)

    def test_env_overrides_opt_in(self) -> None:
        # Callers that want git's "env wins" behaviour prepend the
        # env_config() result themselves.
        with tempfile.NamedTemporaryFile(
            suffix=".gitconfig", delete=False
        ) as global_config:
            global_config.write(b"[user]\n\tname = FromFile\n")
            global_config.flush()
            config_path = global_config.name

        try:
            self.overrideEnv("GIT_CONFIG_GLOBAL", config_path)
            self.overrideEnv("GIT_CONFIG_NOSYSTEM", "1")
            environ = {
                "GIT_CONFIG_COUNT": "1",
                "GIT_CONFIG_KEY_0": "user.name",
                "GIT_CONFIG_VALUE_0": "FromEnv",
            }
            backends = StackedConfig.default_backends()
            override = env_config(environ)
            self.assertIsNotNone(override)
            sc = StackedConfig([override, *backends])
            self.assertEqual(b"FromEnv", sc.get((b"user",), b"name"))
        finally:
            os.unlink(config_path)


class EscapeValueTests(TestCase):
    def test_nothing(self) -> None:
        self.assertEqual(b"foo", _escape_value(b"foo"))

    def test_backslash(self) -> None:
        self.assertEqual(b"foo\\\\", _escape_value(b"foo\\"))

    def test_newline(self) -> None:
        self.assertEqual(b"foo\\n", _escape_value(b"foo\n"))


class FormatStringTests(TestCase):
    def test_quoted(self) -> None:
        self.assertEqual(b'" foo"', _format_string(b" foo"))
        self.assertEqual(b'"\\tfoo"', _format_string(b"\tfoo"))

    def test_not_quoted(self) -> None:
        self.assertEqual(b"foo", _format_string(b"foo"))
        self.assertEqual(b"foo bar", _format_string(b"foo bar"))


class ParseStringTests(TestCase):
    def test_quoted(self) -> None:
        self.assertEqual(b" foo", _parse_string(b'" foo"'))
        self.assertEqual(b"foo ", _parse_string(b'"foo "'))
        self.assertEqual(b"\tfoo", _parse_string(b'"\\tfoo"'))
        self.assertEqual(b"foo\t", _parse_string(b'"foo\\t"'))

    def test_not_quoted(self) -> None:
        self.assertEqual(b"foo", _parse_string(b"foo"))
        self.assertEqual(b"foo bar", _parse_string(b"foo bar"))

    def test_nothing(self) -> None:
        self.assertEqual(b"", _parse_string(b""))

    def test_tab(self) -> None:
        self.assertEqual(b"\tbar\t", _parse_string(b"\\tbar\\t"))

    def test_newline(self) -> None:
        self.assertEqual(b"\nbar\t", _parse_string(b"\\nbar\\t\t"))

    def test_quote(self) -> None:
        self.assertEqual(b'"foo"', _parse_string(b'\\"foo\\"'))


class CheckVariableNameTests(TestCase):
    def test_invalid(self) -> None:
        self.assertFalse(_check_variable_name(b"foo "))
        self.assertFalse(_check_variable_name(b"bar,bar"))
        self.assertFalse(_check_variable_name(b"bar.bar"))

    def test_valid(self) -> None:
        self.assertTrue(_check_variable_name(b"FOO"))
        self.assertTrue(_check_variable_name(b"foo"))
        self.assertTrue(_check_variable_name(b"foo-bar"))


class CheckSectionNameTests(TestCase):
    def test_invalid(self) -> None:
        self.assertFalse(_check_section_name(b"foo "))
        self.assertFalse(_check_section_name(b"bar,bar"))

    def test_valid(self) -> None:
        self.assertTrue(_check_section_name(b"FOO"))
        self.assertTrue(_check_section_name(b"foo"))
        self.assertTrue(_check_section_name(b"foo-bar"))
        self.assertTrue(_check_section_name(b"bar.bar"))


class SubmodulesTests(TestCase):
    def testSubmodules(self) -> None:
        cf = ConfigFile.from_file(
            BytesIO(
                b"""\
[submodule "core/lib"]
\tpath = core/lib
\turl = https://github.com/phhusson/QuasselC.git
"""
            )
        )
        got = list(parse_submodules(cf))
        self.assertEqual(
            [
                (
                    b"core/lib",
                    b"https://github.com/phhusson/QuasselC.git",
                    b"core/lib",
                )
            ],
            got,
        )

    def testMalformedSubmodules(self) -> None:
        cf = ConfigFile.from_file(
            BytesIO(
                b"""\
[submodule "core/lib"]
\tpath = core/lib
\turl = https://github.com/phhusson/QuasselC.git

[submodule "dulwich"]
\turl = https://github.com/jelmer/dulwich
"""
            )
        )
        got = list(parse_submodules(cf))
        self.assertEqual(
            [
                (
                    b"core/lib",
                    b"https://github.com/phhusson/QuasselC.git",
                    b"core/lib",
                )
            ],
            got,
        )

    def test_parse_submodules_skips_sectionless_include(self) -> None:
        cf = ConfigFile.from_file(
            BytesIO(
                b'[submodule "a"]\n\tpath = a\n\turl = ./a\n[include]\n\tpath = x\n'
            ),
            expand_includes=False,
        )
        self.assertEqual([(b"a", b"./a", b"a")], list(parse_submodules(cf)))

    def test_read_submodules_ignores_includes(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = os.path.realpath(tmpdir)
            with open(os.path.join(tmpdir, "evil.config"), "wb") as f:
                f.write(b'[submodule "leaked"]\n\tpath = p\n\turl = file:///\n')
            gitmodules = os.path.join(tmpdir, ".gitmodules")
            with open(gitmodules, "wb") as f:
                f.write(
                    b'[submodule "real"]\n\tpath = sub\n\turl = ./sub\n'
                    b"[include]\n\tpath = evil.config\n"
                )
            self.assertEqual(
                [(b"sub", b"./sub", b"real")], list(read_submodules(gitmodules))
            )


class ApplyInsteadOfTests(TestCase):
    def test_none(self) -> None:
        config = ConfigDict()
        self.assertEqual(
            "https://example.com/", apply_instead_of(config, "https://example.com/")
        )

    def test_apply(self) -> None:
        config = ConfigDict()
        config.set(("url", "https://samba.org/"), "insteadOf", "https://example.com/")
        self.assertEqual(
            "https://samba.org/", apply_instead_of(config, "https://example.com/")
        )

    def test_apply_multiple(self) -> None:
        config = ConfigDict()
        config.add(("url", "https://samba.org/"), "insteadOf", "https://blah.com/")
        config.add(("url", "https://samba.org/"), "insteadOf", "https://example.com/")
        self.assertEqual(
            [b"https://blah.com/", b"https://example.com/"],
            list(config.get_multivar(("url", "https://samba.org/"), "insteadOf")),
        )
        self.assertEqual(
            "https://samba.org/", apply_instead_of(config, "https://example.com/")
        )

    def test_apply_preserves_case_in_subsection(self) -> None:
        """Test that mixed-case URLs (like those with access tokens) are preserved."""
        config = ConfigDict()
        # GitHub access tokens have mixed case that must be preserved
        url_with_token = "https://ghp_AbCdEfGhIjKlMnOpQrStUvWxYz1234567890@github.com/"
        config.set(("url", url_with_token), "insteadOf", "https://github.com/")

        # Apply the substitution
        result = apply_instead_of(config, "https://github.com/jelmer/dulwich.git")
        expected = "https://ghp_AbCdEfGhIjKlMnOpQrStUvWxYz1234567890@github.com/jelmer/dulwich.git"
        self.assertEqual(expected, result)

        # Verify the token case is preserved
        self.assertIn("ghp_AbCdEfGhIjKlMnOpQrStUvWxYz1234567890", result)


class CaseInsensitiveConfigTests(TestCase):
    def test_case_insensitive(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value"
        self.assertEqual("value", config[("CORE",)])
        self.assertEqual("value", config[("CoRe",)])
        self.assertEqual([("core",)], list(config.keys()))

    def test_multiple_set(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value1"
        config[("core",)] = "value2"
        # The second set overwrites the first one
        self.assertEqual("value2", config[("core",)])
        self.assertEqual("value2", config[("CORE",)])

    def test_get_all(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value1"
        config[("CORE",)] = "value2"
        config[("CoRe",)] = "value3"
        self.assertEqual(
            ["value1", "value2", "value3"], list(config.get_all(("core",)))
        )
        self.assertEqual(
            ["value1", "value2", "value3"], list(config.get_all(("CORE",)))
        )

    def test_delitem(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value1"
        config[("CORE",)] = "value2"
        config[("other",)] = "value3"
        del config[("core",)]
        self.assertNotIn(("core",), config)
        self.assertNotIn(("CORE",), config)
        self.assertEqual("value3", config[("other",)])
        self.assertEqual(1, len(config))

    def test_len(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        self.assertEqual(0, len(config))
        config[("core",)] = "value1"
        self.assertEqual(1, len(config))
        config[("CORE",)] = "value2"
        self.assertEqual(1, len(config))  # Same key, case insensitive

    def test_subsection_case_preserved(self) -> None:
        """Test that subsection names preserve their case."""
        config = CaseInsensitiveOrderedMultiDict()
        # Section names should be case-insensitive, but subsection names should preserve case
        config[("url", "https://Example.COM/Path")] = "value1"

        # Can retrieve with different case section name
        self.assertEqual("value1", config[("URL", "https://Example.COM/Path")])
        self.assertEqual("value1", config[("url", "https://Example.COM/Path")])

        # But not with different case subsection name
        with self.assertRaises(KeyError):
            config[("url", "https://example.com/path")]

        # Verify the stored key preserves subsection case
        stored_keys = list(config.keys())
        self.assertEqual(1, len(stored_keys))
        self.assertEqual(("url", "https://Example.COM/Path"), stored_keys[0])
        config[("other",)] = "value3"
        self.assertEqual(2, len(config))

    def test_make_from_dict(self) -> None:
        original = {("core",): "value1", ("other",): "value2"}
        config = CaseInsensitiveOrderedMultiDict.make(original)
        self.assertEqual("value1", config[("core",)])
        self.assertEqual("value1", config[("CORE",)])
        self.assertEqual("value2", config[("other",)])

    def test_make_from_self(self) -> None:
        config1 = CaseInsensitiveOrderedMultiDict()
        config1[("core",)] = "value"
        config2 = CaseInsensitiveOrderedMultiDict.make(config1)
        self.assertIs(config1, config2)

    def test_make_invalid_type(self) -> None:
        self.assertRaises(TypeError, CaseInsensitiveOrderedMultiDict.make, "invalid")

    def test_get_with_default(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value"
        self.assertEqual("value", config.get(("core",)))
        self.assertEqual("value", config.get(("CORE",)))
        self.assertEqual("default", config.get(("missing",), "default"))
        # Test default_factory behavior
        config_with_factory = CaseInsensitiveOrderedMultiDict(
            default_factory=CaseInsensitiveOrderedMultiDict
        )
        result = config_with_factory.get(("missing",))
        self.assertIsInstance(result, CaseInsensitiveOrderedMultiDict)
        self.assertEqual(0, len(result))

    def test_setdefault(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        # Set new value
        result1 = config.setdefault(("core",), "value1")
        self.assertEqual("value1", result1)
        self.assertEqual("value1", config[("core",)])
        # Try to set again with different case - should return existing
        result2 = config.setdefault(("CORE",), "value2")
        self.assertEqual("value1", result2)
        self.assertEqual("value1", config[("core",)])

    def test_values(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value1"
        config[("other",)] = "value2"
        config[("CORE",)] = "value3"  # Overwrites previous core value
        self.assertEqual({"value3", "value2"}, set(config.values()))

    def test_items_iteration(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("core",)] = "value1"
        config[("other",)] = "value2"
        config[("CORE",)] = "value3"
        items = list(config.items())
        self.assertEqual(3, len(items))
        self.assertEqual((("core",), "value1"), items[0])
        self.assertEqual((("other",), "value2"), items[1])
        self.assertEqual((("CORE",), "value3"), items[2])

    def test_str_keys(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config["core"] = "value"
        self.assertEqual("value", config["CORE"])
        self.assertEqual("value", config["CoRe"])

    def test_nested_tuple_keys(self) -> None:
        config = CaseInsensitiveOrderedMultiDict()
        config[("branch", "master")] = "value"
        # Section names are case-insensitive
        self.assertEqual("value", config[("BRANCH", "master")])
        self.assertEqual("value", config[("Branch", "master")])
        # But subsection names are case-sensitive
        with self.assertRaises(KeyError):
            config[("branch", "MASTER")]


class GetGitProxyCommandTests(TestCase):
    def test_no_proxy_configured(self) -> None:
        config = ConfigDict()
        self.assertIsNone(get_git_proxy_command(config, "example.com"))

    def test_default_proxy(self) -> None:
        config = ConfigDict()
        config.set((b"core",), b"gitProxy", b"my-proxy")
        self.assertEqual("my-proxy", get_git_proxy_command(config, "example.com"))

    def test_domain_specific_proxy(self) -> None:
        config = ConfigFile.from_file(
            BytesIO(
                b"[core]\n"
                b"\tgitProxy = proxy-for-kernel for kernel.org\n"
                b"\tgitProxy = default-proxy\n"
            )
        )
        self.assertEqual(
            "proxy-for-kernel", get_git_proxy_command(config, "kernel.org")
        )
        self.assertEqual(
            "proxy-for-kernel", get_git_proxy_command(config, "git.kernel.org")
        )
        self.assertEqual("default-proxy", get_git_proxy_command(config, "example.com"))

    def test_no_default_no_match(self) -> None:
        config = ConfigFile.from_file(
            BytesIO(b"[core]\n\tgitProxy = proxy-cmd for kernel.org\n")
        )
        self.assertIsNone(get_git_proxy_command(config, "example.com"))

    def test_exact_domain_match(self) -> None:
        config = ConfigFile.from_file(
            BytesIO(b"[core]\n\tgitProxy = proxy-cmd for org\n")
        )
        # "kernel.org" ends with ".org", so it matches
        self.assertEqual("proxy-cmd", get_git_proxy_command(config, "kernel.org"))
        # "org" is an exact match
        self.assertEqual("proxy-cmd", get_git_proxy_command(config, "org"))
        # "notorg" should not match
        self.assertIsNone(get_git_proxy_command(config, "notorg"))


class ConfigFileSetTests(TestCase):
    def test_set_replaces_value(self) -> None:
        # Test that set() replaces the value instead of appending
        cf = ConfigFile()
        cf.set((b"core",), b"sshCommand", b"ssh -i ~/.ssh/id_rsa1")
        cf.set((b"core",), b"sshCommand", b"ssh -i ~/.ssh/id_rsa2")

        # Should only have one value
        self.assertEqual(b"ssh -i ~/.ssh/id_rsa2", cf.get((b"core",), b"sshCommand"))

        # When written to file, should only have one entry
        f = BytesIO()
        cf.write_to_file(f)
        content = f.getvalue()
        self.assertEqual(1, content.count(b"sshCommand"))
        self.assertIn(b"sshCommand = ssh -i ~/.ssh/id_rsa2", content)
        self.assertNotIn(b"id_rsa1", content)



# release_robot.py
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for release_robot."""

import datetime
import logging
import os
import re
import shutil
import sys
import tempfile
import time
import unittest
from typing import ClassVar
from unittest.mock import MagicMock, patch

from dulwich.objects import Commit, Tag
from dulwich.repo import Repo
from dulwich.tests.utils import make_commit, make_tag

from . import release_robot

BASEDIR = os.path.abspath(os.path.dirname(__file__))  # this directory


def gmtime_to_datetime(gmt):
    return datetime.datetime(*time.gmtime(gmt)[:6])


class TagPatternTests(unittest.TestCase):
    """test tag patterns."""

    def test_tag_pattern(self) -> None:
        """Test tag patterns."""
        test_cases = {
            "0.3": "0.3",
            "v0.3": "0.3",
            "release0.3": "0.3",
            "Release-0.3": "0.3",
            "v0.3rc1": "0.3rc1",
            "v0.3-rc1": "0.3-rc1",
            "v0.3-rc.1": "0.3-rc.1",
            "version 0.3": "0.3",
            "version_0.3_rc_1": "0.3_rc_1",
            "v1": "1",
            "0.3rc1": "0.3rc1",
        }
        for testcase, version in test_cases.items():
            matches = re.match(release_robot.PATTERN, testcase)
            assert matches is not None
            self.assertEqual(matches.group(1), version)

    def test_pattern_no_match(self) -> None:
        """Test tags that don't match the pattern."""
        test_cases = [
            "master",
            "HEAD",
            "feature-branch",
            "no-numbers",
            "_",
        ]
        for testcase in test_cases:
            matches = re.match(release_robot.PATTERN, testcase)
            self.assertIsNone(matches)


class GetRecentTagsTest(unittest.TestCase):
    """test get recent tags."""

    # Git repo for dulwich project
    test_repo = os.path.join(BASEDIR, "dulwich_test_repo.zip")
    committer = b"Mark Mikofski <mark.mikofski@sunpowercorp.com>"
    test_tags: ClassVar[list[bytes]] = [b"v0.1a", b"v0.1"]
    # Commit times and tag times per tag. Object ids are filled in by
    # setUpClass once the objects have been created, since the contents (and
    # thus the ids) are computed from the commit/tag data.
    tag_test_data: ClassVar[dict[bytes, list]] = {
        test_tags[0]: [1484788003, None, None],
        test_tags[1]: [1484788314, None, [1484788401, None]],
    }

    # Class attributes set in setUpClass
    projdir: ClassVar[str]
    repo: ClassVar[Repo]
    c1: ClassVar[Commit]
    c2: ClassVar[Commit]
    t1: ClassVar[bytes]
    t2: ClassVar[Tag]

    @classmethod
    def setUpClass(cls) -> None:
        cls.projdir = tempfile.mkdtemp()  # temporary project directory
        cls.repo = Repo.init(cls.projdir)  # test repo
        obj_store = cls.repo.object_store  # test repo object store
        # commit 1 ('2017-01-19T01:06:43')
        cls.c1 = make_commit(
            commit_time=cls.tag_test_data[cls.test_tags[0]][0],
            message=b"unannotated tag",
            author=cls.committer,
        )
        obj_store.add_object(cls.c1)
        cls.tag_test_data[cls.test_tags[0]][1] = cls.c1.id
        # tag 1: unannotated
        cls.t1 = cls.test_tags[0]
        cls.repo[b"refs/tags/" + cls.t1] = cls.c1.id  # add unannotated tag
        # commit 2 ('2017-01-19T01:11:54')
        cls.c2 = make_commit(
            commit_time=cls.tag_test_data[cls.test_tags[1]][0],
            message=b"annotated tag",
            parents=[cls.c1.id],
            author=cls.committer,
        )
        obj_store.add_object(cls.c2)
        cls.tag_test_data[cls.test_tags[1]][1] = cls.c2.id
        # tag 2: annotated ('2017-01-19T01:13:21')
        tag_data = cls.tag_test_data[cls.test_tags[1]][2]
        if tag_data is None:
            raise AssertionError("test_tags[1] should have annotated tag data")
        cls.t2 = make_tag(
            cls.c2,
            name=cls.test_tags[1],
            tag_time=tag_data[0],
        )
        obj_store.add_object(cls.t2)
        tag_data[1] = cls.t2.id
        cls.repo[b"refs/heads/master"] = cls.c2.id
        cls.repo[b"refs/tags/" + cls.t2.name] = cls.t2.id  # add annotated tag

    @classmethod
    def tearDownClass(cls) -> None:
        cls.repo.close()
        shutil.rmtree(cls.projdir)

    def test_get_recent_tags(self) -> None:
        """Test get recent tags."""
        tags = release_robot.get_recent_tags(self.projdir)  # get test tags
        for tag, metadata in tags:
            tag_bytes = tag.encode("utf-8")
            test_data = self.tag_test_data[tag_bytes]  # test data tag
            # test commit date, id and author name
            self.assertEqual(metadata[0], gmtime_to_datetime(test_data[0]))
            self.assertEqual(metadata[1].encode("utf-8"), test_data[1])
            self.assertEqual(metadata[2].encode("utf-8"), self.committer)
            # skip unannotated tags
            tag_obj = test_data[2]
            if not tag_obj:
                continue
            # tag date, id and name
            self.assertEqual(metadata[3][0], gmtime_to_datetime(tag_obj[0]))
            self.assertEqual(metadata[3][1].encode("utf-8"), tag_obj[1])
            self.assertEqual(metadata[3][2], tag)

    def test_get_recent_tags_sorting(self) -> None:
        """Test that tags are sorted by commit time from newest to oldest."""
        tags = release_robot.get_recent_tags(self.projdir)
        # v0.1 should be first as it's newer
        self.assertEqual(tags[0][0], "v0.1")
        # v0.1a should be second as it's older
        self.assertEqual(tags[1][0], "v0.1a")

    def test_get_recent_tags_non_tag_refs(self) -> None:
        """Test that non-tag refs are ignored."""
        # Create a commit on a branch to test that it's not included
        branch_commit = make_commit(
            message=b"branch commit",
            author=self.committer,
            commit_time=int(time.time()),
        )
        self.repo.object_store.add_object(branch_commit)
        self.repo[b"refs/heads/test-branch"] = branch_commit.id

        # Get tags and ensure only the actual tags are returned
        tags = release_robot.get_recent_tags(self.projdir)
        self.assertEqual(len(tags), 2)  # Still only 2 tags
        tag_names = [tag[0] for tag in tags]
        self.assertIn("v0.1", tag_names)
        self.assertIn("v0.1a", tag_names)
        self.assertNotIn("test-branch", tag_names)


class GetCurrentVersionTests(unittest.TestCase):
    """Test get_current_version function."""

    def setUp(self):
        """Set up a test repository for each test."""
        self.projdir = tempfile.mkdtemp()
        self.repo = Repo.init(self.projdir)
        self.addCleanup(self.cleanup)

    def cleanup(self):
        """Clean up after test."""
        self.repo.close()
        shutil.rmtree(self.projdir)

    def test_no_tags(self):
        """Test behavior when repo has no tags."""
        # Create a repo with no tags
        result = release_robot.get_current_version(self.projdir)
        self.assertIsNone(result)

    def test_tag_with_pattern_match(self):
        """Test with a tag that matches the pattern."""
        # Create a test commit and tag
        c = make_commit(message=b"Test commit")
        self.repo.object_store.add_object(c)
        self.repo[b"refs/tags/v1.2.3"] = c.id
        self.repo[b"HEAD"] = c.id

        # Test that the version is extracted correctly
        result = release_robot.get_current_version(self.projdir)
        self.assertEqual("1.2.3", result)

    def test_tag_no_pattern_match(self):
        """Test with a tag that doesn't match the pattern."""
        # Create a test commit and tag that won't match the default pattern
        c = make_commit(message=b"Test commit")
        self.repo.object_store.add_object(c)
        self.repo[b"refs/tags/no-version-tag"] = c.id
        self.repo[b"HEAD"] = c.id

        # Test that the full tag is returned when no match
        result = release_robot.get_current_version(self.projdir)
        self.assertEqual("no-version-tag", result)

    def test_with_logger(self):
        """Test with a logger when regex match fails."""
        # Create a test commit and tag that won't match the pattern
        c = make_commit(message=b"Test commit")
        self.repo.object_store.add_object(c)
        self.repo[b"refs/tags/no-version-tag"] = c.id
        self.repo[b"HEAD"] = c.id

        # Create a logger
        logger = logging.getLogger("test_logger")

        # Test with the logger
        result = release_robot.get_current_version(self.projdir, logger=logger)
        self.assertEqual("no-version-tag", result)

    def test_custom_pattern(self):
        """Test with a custom regex pattern."""
        # Create a test commit and tag
        c = make_commit(message=b"Test commit")
        self.repo.object_store.add_object(c)
        self.repo[b"refs/tags/CUSTOM-99.88.77"] = c.id
        self.repo[b"HEAD"] = c.id

        # Test with a custom pattern
        custom_pattern = r"CUSTOM-([\d\.]+)"
        result = release_robot.get_current_version(self.projdir, pattern=custom_pattern)
        self.assertEqual("99.88.77", result)

    def test_with_logger_debug_call(self):
        """Test that the logger.debug method is actually called."""
        # Create a test commit and tag that won't match the pattern
        c = make_commit(message=b"Test commit")
        self.repo.object_store.add_object(c)
        self.repo[b"refs/tags/no-version-tag"] = c.id
        self.repo[b"HEAD"] = c.id

        # Create a mock logger
        mock_logger = MagicMock()

        # Test with the mock logger
        result = release_robot.get_current_version(self.projdir, logger=mock_logger)

        # Verify logger.debug was called
        mock_logger.debug.assert_called_once()
        # Check the tag name is in the debug message
        self.assertIn("no-version-tag", mock_logger.debug.call_args[0][2])

        # The result should still be the full tag
        self.assertEqual("no-version-tag", result)

    def test_multiple_tags(self):
        """Test behavior with multiple tags to ensure we get the most recent."""
        # Create multiple commits and tags with different timestamps
        c1 = make_commit(message=b"First commit", commit_time=1000)
        c2 = make_commit(message=b"Second commit", commit_time=2000, parents=[c1.id])

        self.repo.object_store.add_object(c1)
        self.repo.object_store.add_object(c2)

        # Add tags with older commit first
        self.repo[b"refs/tags/v0.9.0"] = c1.id
        self.repo[b"refs/tags/v1.0.0"] = c2.id
        self.repo[b"HEAD"] = c2.id

        # Get the current version - should be from the most recent commit
        result = release_robot.get_current_version(self.projdir)
        self.assertEqual("1.0.0", result)


class MainFunctionTests(unittest.TestCase):
    """Test the __main__ block."""

    def setUp(self):
        """Set up a test repository."""
        self.projdir = tempfile.mkdtemp()
        self.repo = Repo.init(self.projdir)
        # Create a test commit and tag
        c = make_commit(message=b"Test commit")
        self.repo.object_store.add_object(c)
        self.repo[b"refs/tags/v3.2.1"] = c.id
        self.repo[b"HEAD"] = c.id
        self.addCleanup(self.cleanup)

    def cleanup(self):
        """Clean up after test."""
        self.repo.close()
        shutil.rmtree(self.projdir)

    @patch.object(sys, "argv", ["release_robot.py"])
    @patch("builtins.print")
    def test_main_default_dir(self, mock_print):
        """Test main function with default directory."""
        # Run the __main__ block code with mocked environment
        module_globals = {
            "__name__": "__main__",
            "sys": sys,
            "get_current_version": lambda projdir: "3.2.1",
            "PROJDIR": ".",
        }
        exec(
            compile(
                "if __name__ == '__main__':\n    if len(sys.argv) > 1:\n        _PROJDIR = sys.argv[1]\n    else:\n        _PROJDIR = PROJDIR\n    print(get_current_version(projdir=_PROJDIR))",
                "<string>",
                "exec",
            ),
            module_globals,
        )

        # Check that print was called with the version
        mock_print.assert_called_once_with("3.2.1")

    @patch.object(sys, "argv", ["release_robot.py", "/custom/path"])
    @patch("builtins.print")
    @patch("contrib.release_robot.get_current_version")
    def test_main_custom_dir(self, mock_get_version, mock_print):
        """Test main function with custom directory from command line."""
        mock_get_version.return_value = "4.5.6"

        # Run the __main__ block code with mocked environment
        module_globals = {
            "__name__": "__main__",
            "sys": sys,
            "get_current_version": mock_get_version,
            "PROJDIR": ".",
        }
        exec(
            compile(
                "if __name__ == '__main__':\n    if len(sys.argv) > 1:\n        _PROJDIR = sys.argv[1]\n    else:\n        _PROJDIR = PROJDIR\n    print(get_current_version(projdir=_PROJDIR))",
                "<string>",
                "exec",
            ),
            module_globals,
        )

        # Check that get_current_version was called with the right arg
        mock_get_version.assert_called_once_with(projdir="/custom/path")
        mock_print.assert_called_once_with("4.5.6")



# test_sha256_pack.py -- Tests for SHA256 pack support
# Copyright (C) 2024 The Dulwich contributors
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as public by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for SHA256 pack support in Dulwich."""

import shutil
import tempfile
import unittest
from io import BytesIO

from dulwich.object_format import SHA256
from dulwich.pack import (
    load_pack_index_file,
    write_pack_index_v2,
)


class SHA256PackTests(unittest.TestCase):
    """Tests for SHA256 pack support."""

    def setUp(self):
        """Set up test repository directory."""
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Clean up test repository."""
        shutil.rmtree(self.test_dir)

    def test_pack_index_v2_with_sha256(self):
        """Test that pack index v2 correctly handles SHA256 hashes."""
        # Create SHA256 entries manually (simulating what would happen in a SHA256 repo)
        entries = []
        for i in range(5):
            # Create a fake SHA256 hash
            sha256_hash = SHA256.hash_func(f"test object {i}".encode()).digest()
            offset = i * 1000  # Fake offsets
            crc32 = i  # Fake CRC32
            entries.append((sha256_hash, offset, crc32))

        # Sort entries by SHA (required for pack index)
        entries.sort(key=lambda e: e[0])

        # Write SHA256 pack index with SHA1 pack checksum (Git always uses SHA1 for pack checksums)
        index_buf = BytesIO()
        from hashlib import sha1

        pack_checksum = sha1(b"fake pack data").digest()
        write_pack_index_v2(index_buf, entries, pack_checksum)

        # Load and verify the index
        index_buf.seek(0)
        pack_idx = load_pack_index_file("<memory>", index_buf, SHA256)

        # Check that the index loaded correctly
        self.assertEqual(len(pack_idx), 5)
        self.assertEqual(pack_idx.version, 2)

        # Verify hash_size detection
        self.assertEqual(pack_idx.hash_size, 32)

        # Verify we can look up objects by SHA256
        for sha256_hash, offset, _ in entries:
            # This should not raise KeyError
            found_offset = pack_idx.object_offset(sha256_hash)
            self.assertEqual(found_offset, offset)

    def test_pack_index_v1_with_sha256(self):
        """Test that pack index v1 correctly rejects SHA256 hashes."""
        # Create SHA256 entries manually
        entries = []
        for i in range(5):
            # Create a fake SHA256 hash
            sha256_hash = SHA256.hash_func(f"test v1 object {i}".encode()).digest()
            offset = i * 1000  # Fake offsets
            crc32 = None  # v1 doesn't store CRC32
            entries.append((sha256_hash, offset, crc32))

        # Sort entries by SHA (required for pack index)
        entries.sort(key=lambda e: e[0])

        # Import write_pack_index_v1
        from dulwich.pack import write_pack_index_v1

        # Write SHA256 pack index v1 with SHA1 pack checksum
        index_buf = BytesIO()
        from hashlib import sha1

        pack_checksum = sha1(b"fake v1 pack data").digest()

        # Pack index v1 only supports SHA-1, so this should raise TypeError
        with self.assertRaises(TypeError) as cm:
            write_pack_index_v1(index_buf, entries, pack_checksum)

        self.assertIn("pack index v1 only supports SHA-1", str(cm.exception))


if __name__ == "__main__":
    unittest.main()



# worktree.py -- Porcelain-like interface for Git worktrees
# Copyright (C) 2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Porcelain-like interface for Git worktrees."""

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..worktree import WorkTreeInfo
    from . import RepoPath


def worktree_list(repo: "RepoPath" = ".") -> list["WorkTreeInfo"]:
    """List all worktrees for a repository.

    Args:
        repo: Path to repository

    Returns:
        List of WorkTreeInfo objects
    """
    from ..worktree import list_worktrees
    from . import open_repo_closing

    with open_repo_closing(repo) as r:
        return list_worktrees(r)


def worktree_add(
    repo: "RepoPath" = ".",
    path: str | os.PathLike[str] | None = None,
    branch: str | bytes | None = None,
    commit: str | bytes | None = None,
    detach: bool = False,
    force: bool = False,
) -> str:
    """Add a new worktree.

    Args:
        repo: Path to repository
        path: Path for new worktree
        branch: Branch to checkout (creates if doesn't exist)
        commit: Specific commit to checkout
        detach: Create with detached HEAD
        force: Force creation even if branch is already checked out

    Returns:
        Path to the newly created worktree
    """
    from ..objects import ObjectID
    from ..worktree import add_worktree
    from . import open_repo_closing

    if path is None:
        raise ValueError("Path is required for worktree add")

    with open_repo_closing(repo) as r:
        commit_bytes = commit.encode() if isinstance(commit, str) else commit
        commit_id = ObjectID(commit_bytes) if commit_bytes is not None else None
        wt_repo = add_worktree(
            r, path, branch=branch, commit=commit_id, detach=detach, force=force
        )
        return wt_repo.path


def worktree_remove(
    repo: "RepoPath" = ".",
    path: str | os.PathLike[str] | None = None,
    force: bool = False,
) -> None:
    """Remove a worktree.

    Args:
        repo: Path to repository
        path: Path to worktree to remove
        force: Force removal even if there are local changes
    """
    from ..worktree import remove_worktree
    from . import open_repo_closing

    if path is None:
        raise ValueError("Path is required for worktree remove")

    with open_repo_closing(repo) as r:
        remove_worktree(r, path, force=force)


def worktree_prune(
    repo: "RepoPath" = ".", dry_run: bool = False, expire: int | None = None
) -> list[str]:
    """Prune worktree administrative files.

    Args:
        repo: Path to repository
        dry_run: Only show what would be removed
        expire: Only prune worktrees older than this many seconds

    Returns:
        List of pruned worktree names
    """
    from ..worktree import prune_worktrees
    from . import open_repo_closing

    with open_repo_closing(repo) as r:
        return prune_worktrees(r, expire=expire, dry_run=dry_run)


def worktree_lock(
    repo: "RepoPath" = ".",
    path: str | os.PathLike[str] | None = None,
    reason: str | None = None,
) -> None:
    """Lock a worktree to prevent it from being pruned.

    Args:
        repo: Path to repository
        path: Path to worktree to lock
        reason: Optional reason for locking
    """
    from ..worktree import lock_worktree
    from . import open_repo_closing

    if path is None:
        raise ValueError("Path is required for worktree lock")

    with open_repo_closing(repo) as r:
        lock_worktree(r, path, reason=reason)


def worktree_unlock(
    repo: "RepoPath" = ".", path: str | os.PathLike[str] | None = None
) -> None:
    """Unlock a worktree.

    Args:
        repo: Path to repository
        path: Path to worktree to unlock
    """
    from ..worktree import unlock_worktree
    from . import open_repo_closing

    if path is None:
        raise ValueError("Path is required for worktree unlock")

    with open_repo_closing(repo) as r:
        unlock_worktree(r, path)


def worktree_move(
    repo: "RepoPath" = ".",
    old_path: str | os.PathLike[str] | None = None,
    new_path: str | os.PathLike[str] | None = None,
) -> None:
    """Move a worktree to a new location.

    Args:
        repo: Path to repository
        old_path: Current path of worktree
        new_path: New path for worktree
    """
    from ..worktree import move_worktree
    from . import open_repo_closing

    if old_path is None or new_path is None:
        raise ValueError("Both old_path and new_path are required for worktree move")

    with open_repo_closing(repo) as r:
        move_worktree(r, old_path, new_path)


def worktree_repair(
    repo: "RepoPath" = ".",
    paths: list[str | os.PathLike[str]] | None = None,
) -> list[str]:
    """Repair worktree administrative files.

    Args:
        repo: Path to repository
        paths: Optional list of worktree paths to repair. If None, repairs
               connections from the main repository to all linked worktrees.

    Returns:
        List of repaired worktree paths
    """
    from ..worktree import repair_worktree
    from . import open_repo_closing

    with open_repo_closing(repo) as r:
        return repair_worktree(r, paths=paths)



# test_range_diff.py -- Tests for porcelain range-diff
# Copyright (C) 2026 Jelmer Vernooĳ <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for the porcelain range-diff functionality."""

import importlib.util
import os
import tempfile
import unittest
from io import BytesIO

from dulwich import porcelain
from dulwich.repo import Repo

from .. import TestCase


@unittest.skipUnless(
    importlib.util.find_spec("munkres") is not None, "munkres not available"
)
class PorcelainRangeDiffTests(TestCase):
    """Tests for the porcelain range_diff function."""

    def setUp(self):
        super().setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(self._cleanup)
        self.repo = Repo.init(self.tmpdir)
        self.addCleanup(self.repo.close)

    def _cleanup(self):
        import shutil

        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def commit(self, message, files):
        for name, content in files.items():
            path = os.path.join(self.tmpdir, name)
            with open(path, "wb") as f:
                f.write(content)
            porcelain.add(self.repo, paths=[path])
        return porcelain.commit(
            self.repo,
            message=message,
            author=b"Test <test@example.com>",
            committer=b"Test <test@example.com>",
        )

    def _build_two_ranges(self):
        base = self.commit(b"base", {"a": b"a\n"})
        old1 = self.commit(b"shared", {"b": b"b\n"})
        old_tip = self.commit(b"dropped", {"c": b"c\n"})
        porcelain.reset(self.repo, mode="hard", treeish=old1)
        new_tip = self.commit(b"added", {"d": b"d\n"})
        return base, old_tip, new_tip

    def test_two_range_form(self):
        base, old_tip, new_tip = self._build_two_ranges()
        out = BytesIO()
        porcelain.range_diff(
            self.repo,
            base + b".." + old_tip,
            base + b".." + new_tip,
            outstream=out,
        )
        text = out.getvalue().decode()
        self.assertIn("shared", text)
        self.assertIn("dropped", text)
        self.assertIn("added", text)

    def test_base_form(self):
        base, old_tip, new_tip = self._build_two_ranges()
        out = BytesIO()
        porcelain.range_diff(self.repo, old_tip, new_tip, base=base, outstream=out)
        text = out.getvalue().decode()
        self.assertIn("shared", text)

    def test_symmetric_form(self):
        _base, old_tip, new_tip = self._build_two_ranges()
        out = BytesIO()
        porcelain.range_diff(self.repo, old_tip + b"..." + new_tip, outstream=out)
        text = out.getvalue().decode()
        self.assertIn("dropped", text)
        self.assertIn("added", text)

    def test_single_revision_requires_symmetric(self):
        base = self.commit(b"base", {"a": b"a\n"})
        out = BytesIO()
        self.assertRaises(
            porcelain.Error,
            porcelain.range_diff,
            self.repo,
            base,
            outstream=out,
        )

    def test_base_form_requires_two_revisions(self):
        base = self.commit(b"base", {"a": b"a\n"})
        out = BytesIO()
        self.assertRaises(
            porcelain.Error,
            porcelain.range_diff,
            self.repo,
            base,
            base=base,
            outstream=out,
        )

    def test_str_arguments(self):
        base, old_tip, new_tip = self._build_two_ranges()
        out = BytesIO()
        porcelain.range_diff(
            self.repo,
            (base + b".." + old_tip).decode(),
            (base + b".." + new_tip).decode(),
            outstream=out,
        )
        self.assertIn(b"shared", out.getvalue())

    def test_commit_object_arguments(self):
        base, old_tip, new_tip = self._build_two_ranges()
        # Pass parsed Commit objects rather than revision strings.
        base_commit = self.repo[base]
        old_commit = self.repo[old_tip]
        new_commit = self.repo[new_tip]
        out = BytesIO()
        porcelain.range_diff(
            self.repo,
            old_commit,
            new_commit,
            base=base_commit,
            outstream=out,
        )
        self.assertIn(b"shared", out.getvalue())



# notes.py -- Git notes handling
# Copyright (C) 2024 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Git notes handling."""

__all__ = [
    "DEFAULT_NOTES_REF",
    "NOTES_REF_PREFIX",
    "Notes",
    "NotesTree",
    "create_notes_tree",
    "get_note_fanout_level",
    "get_note_path",
    "split_path_for_fanout",
]

import stat
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING

from .objects import Blob, ObjectID, Tree
from .refs import Ref

if TYPE_CHECKING:
    from .config import StackedConfig
    from .object_store import BaseObjectStore
    from .refs import RefsContainer

NOTES_REF_PREFIX = b"refs/notes/"
DEFAULT_NOTES_REF = NOTES_REF_PREFIX + b"commits"


def get_note_fanout_level(tree: Tree, object_store: "BaseObjectStore") -> int:
    """Determine the fanout level for a note tree.

    Git uses a fanout directory structure for performance with large numbers
    of notes. The fanout level determines how many levels of subdirectories
    are used.

    Args:
        tree: The notes tree to analyze
        object_store: Object store to retrieve subtrees

    Returns:
        Fanout level (0 for no fanout, 1 or 2 for fanout)
    """

    # Count the total number of notes in the tree recursively
    def count_notes(tree: Tree, level: int = 0) -> int:
        """Count notes in a tree recursively.

        Args:
            tree: Tree to count notes in
            level: Current recursion level

        Returns:
            Total number of notes
        """
        count = 0
        for name, mode, sha in tree.items():
            assert mode is not None
            if stat.S_ISREG(mode):
                count += 1
            elif stat.S_ISDIR(mode) and level < 2:  # Only recurse 2 levels deep
                assert sha is not None
                try:
                    subtree = object_store[sha]
                    assert isinstance(subtree, Tree)
                    count += count_notes(subtree, level + 1)
                except KeyError:
                    pass
        return count

    note_count = count_notes(tree)

    # Use fanout based on number of notes
    # Git typically starts using fanout around 256 notes
    if note_count < 256:
        return 0
    elif note_count < 65536:  # 256^2
        return 1
    else:
        return 2


def split_path_for_fanout(hexsha: bytes, fanout_level: int) -> tuple[bytes, ...]:
    """Split a hex SHA into path components based on fanout level.

    Args:
        hexsha: Hex SHA of the object
        fanout_level: Number of directory levels for fanout

    Returns:
        Tuple of path components
    """
    if fanout_level == 0:
        return (hexsha,)

    components = []
    for i in range(fanout_level):
        components.append(hexsha[i * 2 : (i + 1) * 2])
    components.append(hexsha[fanout_level * 2 :])
    return tuple(components)


def get_note_path(object_sha: bytes, fanout_level: int = 0) -> bytes:
    """Get the path within the notes tree for a given object.

    Args:
        object_sha: Hex SHA of the object to get notes for
        fanout_level: Fanout level to use

    Returns:
        Path within the notes tree
    """
    components = split_path_for_fanout(object_sha, fanout_level)
    return b"/".join(components)


class NotesTree:
    """Represents a Git notes tree."""

    def __init__(self, tree: Tree, object_store: "BaseObjectStore") -> None:
        """Initialize a notes tree.

        Args:
            tree: The tree object containing notes
            object_store: Object store to retrieve note contents from
        """
        self._tree = tree
        self._object_store = object_store
        self._fanout_level = self._detect_fanout_level()

    def _detect_fanout_level(self) -> int:
        """Detect the fanout level used in this notes tree.

        Returns:
            Detected fanout level
        """
        if not self._tree.items():
            return 0

        # Check for presence of both files and directories
        has_files = False
        has_dirs = False
        dir_names = []

        for name, mode, sha in self._tree.items():
            assert name is not None
            assert mode is not None
            if stat.S_ISDIR(mode):
                has_dirs = True
                dir_names.append(name)
            elif stat.S_ISREG(mode):
                has_files = True

        # If we have files at the root level, check if they're full SHA names
        if has_files and not has_dirs:
            # Check if any file names are full hex strings (40 for SHA-1, 64 for SHA-256)
            hex_length = self._object_store.object_format.hex_length
            for name, mode, sha in self._tree.items():
                assert name is not None
                assert mode is not None
                if stat.S_ISREG(mode) and len(name) == hex_length:
                    try:
                        int(name, 16)  # Verify it's a valid hex string
                        return 0  # No fanout
                    except ValueError:
                        pass

        # Check if all directories are 2-character hex names
        if has_dirs and dir_names:
            all_two_char_hex = all(
                len(name) == 2 and all(c in b"0123456789abcdef" for c in name)
                for name in dir_names
            )

            if all_two_char_hex:
                # Check a sample directory to determine if it's level 1 or 2
                sample_dir_name = dir_names[0]
                try:
                    _sample_mode, sample_sha = self._tree[sample_dir_name]
                    sample_tree = self._object_store[sample_sha]
                    assert isinstance(sample_tree, Tree)

                    # Check if this subtree also has 2-char hex directories
                    sub_has_dirs = False
                    for sub_name, sub_mode, sub_sha in sample_tree.items():
                        assert sub_name is not None
                        assert sub_mode is not None
                        if stat.S_ISDIR(sub_mode) and len(sub_name) == 2:
                            try:
                                int(sub_name, 16)
                                sub_has_dirs = True
                                break
                            except ValueError:
                                pass

                    return 2 if sub_has_dirs else 1
                except KeyError:
                    return 1  # Assume level 1 if we can't inspect

        return 0

    def _reorganize_tree(self, new_fanout_level: int) -> None:
        """Reorganize the notes tree to use a different fanout level.

        Args:
            new_fanout_level: The desired fanout level
        """
        if new_fanout_level == self._fanout_level:
            return

        # Collect all existing notes
        notes = []
        for object_sha, note_sha in self.list_notes():
            note_obj = self._object_store[note_sha]
            if isinstance(note_obj, Blob):
                notes.append((object_sha, note_obj.data))

        # Create new empty tree
        new_tree = Tree()
        self._object_store.add_object(new_tree)
        self._tree = new_tree
        self._fanout_level = new_fanout_level

        # Re-add all notes with new fanout structure using set_note
        # Temporarily set fanout back to avoid recursion
        for object_sha, note_content in notes:
            # Use the internal tree update logic without checking fanout again
            note_blob = Blob.from_string(note_content)
            self._object_store.add_object(note_blob)

            path = get_note_path(object_sha, new_fanout_level)
            components = path.split(b"/")

            # Build new tree structure
            def update_tree(
                tree: Tree, components: Sequence[bytes], blob_sha: ObjectID
            ) -> Tree:
                """Update tree with new note entry.

                Args:
                    tree: Tree to update
                    components: Path components
                    blob_sha: SHA of the note blob

                Returns:
                    Updated tree
                """
                if len(components) == 1:
                    # Leaf level - add the note blob
                    new_tree = Tree()
                    for name, mode, sha in tree.items():
                        if name != components[0]:
                            assert name is not None
                            assert mode is not None
                            assert sha is not None
                            new_tree.add(name, mode, sha)
                    new_tree.add(components[0], stat.S_IFREG | 0o644, blob_sha)
                    return new_tree
                else:
                    # Directory level
                    new_tree = Tree()
                    found = False
                    for name, mode, sha in tree.items():
                        if name == components[0]:
                            # Update this subtree
                            assert mode is not None and sha is not None
                            if stat.S_ISDIR(mode):
                                subtree = self._object_store[sha]
                                assert isinstance(subtree, Tree)
                            else:
                                # If not a directory, we need to replace it
                                subtree = Tree()
                            new_subtree = update_tree(subtree, components[1:], blob_sha)
                            self._object_store.add_object(new_subtree)
                            new_tree.add(name, stat.S_IFDIR, new_subtree.id)
                            found = True
                        else:
                            assert (
                                name is not None
                                and mode is not None
                                and sha is not None
                            )
                            new_tree.add(name, mode, sha)

                    if not found:
                        # Create new subtree path
                        subtree = Tree()
                        new_subtree = update_tree(subtree, components[1:], blob_sha)
                        self._object_store.add_object(new_subtree)
                        new_tree.add(components[0], stat.S_IFDIR, new_subtree.id)

                    return new_tree

            self._tree = update_tree(self._tree, components, note_blob.id)
            self._object_store.add_object(self._tree)

    def _update_tree_entry(
        self, tree: Tree, name: bytes, mode: int, sha: ObjectID
    ) -> Tree:
        """Update a tree entry and return the updated tree.

        Args:
            tree: The tree to update
            name: Name of the entry
            mode: File mode
            sha: SHA of the object

        Returns:
            The updated tree
        """
        new_tree = Tree()
        for existing_name, existing_mode, existing_sha in tree.items():
            if existing_name != name:
                assert (
                    existing_name is not None
                    and existing_mode is not None
                    and existing_sha is not None
                )
                new_tree.add(existing_name, existing_mode, existing_sha)
        new_tree.add(name, mode, sha)
        self._object_store.add_object(new_tree)

        # Update the tree reference
        if tree is self._tree:
            self._tree = new_tree

        return new_tree

    def _get_note_sha(self, object_sha: bytes) -> ObjectID | None:
        """Get the SHA of the note blob for an object.

        Args:
            object_sha: SHA of the object to get notes for

        Returns:
            SHA of the note blob, or None if no note exists
        """
        path = get_note_path(object_sha, self._fanout_level)
        components = path.split(b"/")

        current_tree = self._tree
        for component in components[:-1]:
            try:
                mode, sha = current_tree[component]
                if not stat.S_ISDIR(mode):  # Not a directory
                    return None
                obj = self._object_store[sha]
                assert isinstance(obj, Tree)
                current_tree = obj
            except KeyError:
                return None

        try:
            mode, sha = current_tree[components[-1]]
            if not stat.S_ISREG(mode):  # Not a regular file
                return None
            return sha
        except KeyError:
            return None

    def get_note(self, object_sha: bytes) -> bytes | None:
        """Get the note content for an object.

        Args:
            object_sha: SHA of the object to get notes for

        Returns:
            Note content as bytes, or None if no note exists
        """
        note_sha = self._get_note_sha(object_sha)
        if note_sha is None:
            return None

        try:
            note_obj = self._object_store[note_sha]
            if not isinstance(note_obj, Blob):
                return None
            data: bytes = note_obj.data
            return data
        except KeyError:
            return None

    def set_note(self, object_sha: bytes, note_content: bytes) -> Tree:
        """Set or update a note for an object.

        Args:
            object_sha: SHA of the object to annotate
            note_content: Content of the note

        Returns:
            New tree object with the note added/updated
        """
        # Create note blob
        note_blob = Blob.from_string(note_content)
        self._object_store.add_object(note_blob)

        # Check if we need to reorganize the tree for better fanout
        desired_fanout = get_note_fanout_level(self._tree, self._object_store)
        if desired_fanout != self._fanout_level:
            self._reorganize_tree(desired_fanout)

        # Get path components
        path = get_note_path(object_sha, self._fanout_level)
        components = path.split(b"/")

        # Build new tree structure
        def update_tree(
            tree: Tree, components: Sequence[bytes], blob_sha: ObjectID
        ) -> Tree:
            """Update tree with new note entry.

            Args:
                tree: Tree to update
                components: Path components
                blob_sha: SHA of the note blob

            Returns:
                Updated tree
            """
            if len(components) == 1:
                # Leaf level - add the note blob
                new_tree = Tree()
                for name, mode, sha in tree.items():
                    if name != components[0]:
                        assert name is not None and mode is not None and sha is not None
                        new_tree.add(name, mode, sha)
                new_tree.add(components[0], stat.S_IFREG | 0o644, blob_sha)
                return new_tree
            else:
                # Directory level
                new_tree = Tree()
                found = False
                for name, mode, sha in tree.items():
                    if name == components[0]:
                        # Update this subtree
                        assert mode is not None and sha is not None
                        if stat.S_ISDIR(mode):
                            subtree = self._object_store[sha]
                            assert isinstance(subtree, Tree)
                        else:
                            # If not a directory, we need to replace it
                            subtree = Tree()
                        new_subtree = update_tree(subtree, components[1:], blob_sha)
                        self._object_store.add_object(new_subtree)
                        new_tree.add(name, stat.S_IFDIR, new_subtree.id)
                        found = True
                    else:
                        assert name is not None and mode is not None and sha is not None
                        new_tree.add(name, mode, sha)

                if not found:
                    # Create new subtree path
                    subtree = Tree()
                    new_subtree = update_tree(subtree, components[1:], blob_sha)
                    self._object_store.add_object(new_subtree)
                    new_tree.add(components[0], stat.S_IFDIR, new_subtree.id)

                return new_tree

        new_tree = update_tree(self._tree, components, note_blob.id)
        self._object_store.add_object(new_tree)
        self._tree = new_tree
        self._fanout_level = self._detect_fanout_level()
        return new_tree

    def remove_note(self, object_sha: bytes) -> Tree | None:
        """Remove a note for an object.

        Args:
            object_sha: SHA of the object to remove notes from

        Returns:
            New tree object with the note removed, or None if no note existed
        """
        if self._get_note_sha(object_sha) is None:
            return None

        # Get path components
        path = get_note_path(object_sha, self._fanout_level)
        components = path.split(b"/")

        # Build new tree structure without the note
        def remove_from_tree(tree: Tree, components: Sequence[bytes]) -> Tree | None:
            """Remove note entry from tree.

            Args:
                tree: Tree to remove from
                components: Path components

            Returns:
                Updated tree or None if empty
            """
            if len(components) == 1:
                # Leaf level - remove the note
                new_tree = Tree()
                found = False
                for name, mode, sha in tree.items():
                    if name != components[0]:
                        assert name is not None and mode is not None and sha is not None
                        new_tree.add(name, mode, sha)
                    else:
                        found = True

                if not found:
                    return None

                # Return None if tree is now empty
                return new_tree if len(new_tree) > 0 else None
            else:
                # Directory level
                new_tree = Tree()
                modified = False
                for name, mode, sha in tree.items():
                    assert name is not None and mode is not None and sha is not None
                    if name == components[0] and stat.S_ISDIR(mode):
                        # Update this subtree
                        subtree = self._object_store[sha]
                        assert isinstance(subtree, Tree)
                        new_subtree = remove_from_tree(subtree, components[1:])
                        if new_subtree is not None:
                            self._object_store.add_object(new_subtree)
                            new_tree.add(name, stat.S_IFDIR, new_subtree.id)
                        modified = True
                    else:
                        new_tree.add(name, mode, sha)

                if not modified:
                    return None

                # Return None if tree is now empty
                return new_tree if len(new_tree) > 0 else None

        new_tree = remove_from_tree(self._tree, components)
        if new_tree is None:
            new_tree = Tree()  # Empty tree

        self._object_store.add_object(new_tree)
        self._tree = new_tree
        self._fanout_level = self._detect_fanout_level()
        return new_tree

    def list_notes(self) -> Iterator[tuple[ObjectID, ObjectID]]:
        """List all notes in this tree.

        Yields:
            Tuples of (object_sha, note_sha)
        """

        def walk_tree(
            tree: Tree, prefix: bytes = b""
        ) -> Iterator[tuple[ObjectID, ObjectID]]:
            """Walk the notes tree recursively.

            Args:
                tree: Tree to walk
                prefix: Path prefix for current level

            Yields:
                Tuples of (object_sha, note_sha)
            """
            for name, mode, sha in tree.items():
                assert name is not None and mode is not None and sha is not None
                if stat.S_ISDIR(mode):  # Directory
                    subtree = self._object_store[sha]
                    assert isinstance(subtree, Tree)
                    yield from walk_tree(subtree, prefix + name)
                elif stat.S_ISREG(mode):  # File
                    # Reconstruct the full hex SHA from the path
                    full_hex = prefix + name
                    yield (ObjectID(full_hex), sha)

        yield from walk_tree(self._tree)


def create_notes_tree(object_store: "BaseObjectStore") -> Tree:
    """Create an empty notes tree.

    Args:
        object_store: Object store to add the tree to

    Returns:
        Empty tree object
    """
    tree = Tree()
    object_store.add_object(tree)
    return tree


class Notes:
    """High-level interface for Git notes operations."""

    def __init__(
        self, object_store: "BaseObjectStore", refs_container: "RefsContainer"
    ) -> None:
        """Initialize Notes.

        Args:
            object_store: Object store to read/write objects
            refs_container: Refs container to read/write refs
        """
        self._object_store = object_store
        self._refs = refs_container

    def get_notes_ref(
        self,
        notes_ref: bytes | None = None,
        config: "StackedConfig | None" = None,
    ) -> Ref:
        """Get the notes reference to use.

        Args:
            notes_ref: The notes ref to use, or None to use the default
            config: Config to read notes.displayRef from

        Returns:
            The notes reference name
        """
        if notes_ref is None:
            if config is not None:
                notes_ref = config.get((b"notes",), b"displayRef")
            if notes_ref is None:
                notes_ref = DEFAULT_NOTES_REF
        return Ref(notes_ref)

    def get_note(
        self,
        object_sha: bytes,
        notes_ref: bytes | None = None,
        config: "StackedConfig | None" = None,
    ) -> bytes | None:
        """Get the note for an object.

        Args:
            object_sha: SHA of the object to get notes for
            notes_ref: The notes ref to use, or None to use the default
            config: Config to read notes.displayRef from

        Returns:
            The note content as bytes, or None if no note exists
        """
        notes_ref = self.get_notes_ref(notes_ref, config)
        try:
            notes_commit_sha = self._refs[notes_ref]
        except KeyError:
            return None

        # Get the commit object
        notes_obj = self._object_store[notes_commit_sha]

        # If it's a commit, get the tree from it
        from .objects import Commit

        if isinstance(notes_obj, Commit):
            notes_tree = self._object_store[notes_obj.tree]
        else:
            # If it's directly a tree (shouldn't happen in normal usage)
            notes_tree = notes_obj

        if not isinstance(notes_tree, Tree):
            return None

        notes_tree_obj = NotesTree(notes_tree, self._object_store)
        return notes_tree_obj.get_note(object_sha)

    def set_note(
        self,
        object_sha: bytes,
        note_content: bytes,
        notes_ref: bytes | None = None,
        author: bytes | None = None,
        committer: bytes | None = None,
        message: bytes | None = None,
        config: "StackedConfig | None" = None,
    ) -> bytes:
        """Set or update a note for an object.

        Args:
            object_sha: SHA of the object to annotate
            note_content: Content of the note
            notes_ref: The notes ref to use, or None to use the default
            author: Author identity (defaults to committer)
            committer: Committer identity (defaults to config)
            message: Commit message for the notes update
            config: Config to read user identity and notes.displayRef from

        Returns:
            SHA of the new notes commit
        """
        import time

        from .objects import Commit
        from .repo import get_user_identity

        notes_ref = self.get_notes_ref(notes_ref, config)

        # Get current notes tree
        try:
            notes_commit_sha = self._refs[notes_ref]
            notes_obj = self._object_store[notes_commit_sha]

            # If it's a commit, get the tree from it
            if isinstance(notes_obj, Commit):
                notes_tree = self._object_store[notes_obj.tree]
            else:
                # If it's directly a tree (shouldn't happen in normal usage)
                notes_tree = notes_obj

            if not isinstance(notes_tree, Tree):
                notes_tree = create_notes_tree(self._object_store)
        except KeyError:
            notes_tree = create_notes_tree(self._object_store)

        # Update notes tree
        notes_tree_obj = NotesTree(notes_tree, self._object_store)
        new_tree = notes_tree_obj.set_note(object_sha, note_content)

        # Create commit
        if committer is None and config is not None:
            committer = get_user_identity(config, kind="COMMITTER")
        if committer is None:
            committer = b"Git User <user@example.com>"
        if author is None:
            author = committer
        if message is None:
            message = b"Notes added by 'git notes add'"

        commit = Commit()
        commit.tree = new_tree.id
        commit.author = author
        commit.committer = committer
        commit.commit_time = commit.author_time = int(time.time())
        commit.commit_timezone = commit.author_timezone = 0
        commit.encoding = b"UTF-8"
        commit.message = message

        # Set parent to previous notes commit if exists
        try:
            parent_sha = self._refs[notes_ref]
            parent = self._object_store[parent_sha]
            if isinstance(parent, Commit):
                commit.parents = [parent_sha]
        except KeyError:
            commit.parents = []

        self._object_store.add_object(commit)
        self._refs[notes_ref] = commit.id

        return commit.id

    def remove_note(
        self,
        object_sha: bytes,
        notes_ref: bytes | None = None,
        author: bytes | None = None,
        committer: bytes | None = None,
        message: bytes | None = None,
        config: "StackedConfig | None" = None,
    ) -> bytes | None:
        """Remove a note for an object.

        Args:
            object_sha: SHA of the object to remove notes from
            notes_ref: The notes ref to use, or None to use the default
            author: Author identity (defaults to committer)
            committer: Committer identity (defaults to config)
            message: Commit message for the notes removal
            config: Config to read user identity and notes.displayRef from

        Returns:
            SHA of the new notes commit, or None if no note existed
        """
        import time

        from .objects import Commit
        from .repo import get_user_identity

        notes_ref = self.get_notes_ref(notes_ref, config)

        # Get current notes tree
        try:
            notes_commit_sha = self._refs[notes_ref]
            notes_obj = self._object_store[notes_commit_sha]

            # If it's a commit, get the tree from it
            if isinstance(notes_obj, Commit):
                notes_tree = self._object_store[notes_obj.tree]
            else:
                # If it's directly a tree (shouldn't happen in normal usage)
                notes_tree = notes_obj

            if not isinstance(notes_tree, Tree):
                return None
        except KeyError:
            return None

        # Remove from notes tree
        notes_tree_obj = NotesTree(notes_tree, self._object_store)
        new_tree = notes_tree_obj.remove_note(object_sha)
        if new_tree is None:
            return None

        # Create commit
        if committer is None and config is not None:
            committer = get_user_identity(config, kind="COMMITTER")
        if committer is None:
            committer = b"Git User <user@example.com>"
        if author is None:
            author = committer
        if message is None:
            message = b"Notes removed by 'git notes remove'"

        commit = Commit()
        commit.tree = new_tree.id
        commit.author = author
        commit.committer = committer
        commit.commit_time = commit.author_time = int(time.time())
        commit.commit_timezone = commit.author_timezone = 0
        commit.encoding = b"UTF-8"
        commit.message = message

        # Set parent to previous notes commit
        parent_sha = self._refs[notes_ref]
        parent = self._object_store[parent_sha]
        if isinstance(parent, Commit):
            commit.parents = [parent_sha]

        self._object_store.add_object(commit)
        self._refs[notes_ref] = commit.id

        return commit.id

    def list_notes(
        self,
        notes_ref: bytes | None = None,
        config: "StackedConfig | None" = None,
    ) -> list[tuple[ObjectID, bytes]]:
        """List all notes in a notes ref.

        Args:
            notes_ref: The notes ref to use, or None to use the default
            config: Config to read notes.displayRef from

        Returns:
            List of tuples of (object_sha, note_content)
        """
        notes_ref = self.get_notes_ref(notes_ref, config)
        try:
            notes_commit_sha = self._refs[notes_ref]
        except KeyError:
            return []

        # Get the commit object
        from .objects import Commit

        notes_obj = self._object_store[notes_commit_sha]

        # If it's a commit, get the tree from it
        if isinstance(notes_obj, Commit):
            notes_tree = self._object_store[notes_obj.tree]
        else:
            # If it's directly a tree (shouldn't happen in normal usage)
            notes_tree = notes_obj

        if not isinstance(notes_tree, Tree):
            return []

        notes_tree_obj = NotesTree(notes_tree, self._object_store)
        result: list[tuple[ObjectID, bytes]] = []
        for object_sha, note_sha in notes_tree_obj.list_notes():
            note_obj = self._object_store[note_sha]
            if isinstance(note_obj, Blob):
                result.append((object_sha, note_obj.data))
        return result



# stash.py
# Copyright (C) 2018 Jelmer Vernooij <jelmer@samba.org>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Stash handling."""

__all__ = [
    "DEFAULT_STASH_REF",
    "CommitKwargs",
    "Stash",
]

import os
from typing import TYPE_CHECKING, TypedDict

from .diff_tree import tree_changes
from .file import GitFile
from .index import (
    IndexEntry,
    InvalidPathError,
    _tree_to_fs_path,
    build_file_from_blob,
    commit_tree,
    get_path_element_validator,
    index_entry_from_stat,
    iter_fresh_objects,
    symlink,
    update_working_tree,
    validate_path,
)
from .object_store import iter_tree_contents
from .objects import S_IFGITLINK, Blob, Commit, ObjectID, TreeEntry
from .reflog import drop_reflog_entry, read_reflog
from .refs import Ref

if TYPE_CHECKING:
    from .config import Config
    from .reflog import Entry
    from .repo import Repo


class CommitKwargs(TypedDict, total=False):
    """Keyword arguments for do_commit."""

    committer: bytes
    author: bytes


DEFAULT_STASH_REF = Ref(b"refs/stash")


class Stash:
    """A Git stash.

    Note that this doesn't currently update the working tree.
    """

    def __init__(self, repo: "Repo", ref: Ref = DEFAULT_STASH_REF) -> None:
        """Initialize Stash.

        Args:
          repo: Repository object
          ref: Stash reference name
        """
        self._ref = ref
        self._repo = repo

    @property
    def _reflog_path(self) -> str:
        return os.path.join(self._repo.commondir(), "logs", os.fsdecode(self._ref))

    def stashes(self) -> list["Entry"]:
        """Get list of stash entries.

        Returns:
          List of stash entries in chronological order
        """
        try:
            with GitFile(self._reflog_path, "rb") as f:
                return list(reversed(list(read_reflog(f))))
        except FileNotFoundError:
            return []

    @classmethod
    def from_repo(cls, repo: "Repo") -> "Stash":
        """Create a new stash from a Repo object."""
        return cls(repo)

    def drop(self, index: int) -> None:
        """Drop entry with specified index."""
        with open(self._reflog_path, "rb+") as f:
            drop_reflog_entry(f, index, rewrite=True)
        if len(self) == 0:
            os.remove(self._reflog_path)
            del self._repo.refs[self._ref]
            return
        if index == 0:
            self._repo.refs[self._ref] = self[0].new_sha

    def pop(self, index: int, *, config: "Config | None" = None) -> "Entry":
        """Pop a stash entry and apply its changes.

        Args:
          index: Index of the stash entry to pop (0 is the most recent)
          config: Repository configuration. If None, falls back to
            ``self._repo.get_config_stack()``.

        Returns:
          The stash entry that was popped
        """
        if config is None:
            config = self._repo.get_config_stack()
        # Get the stash entry before removing it
        entry = self[index]

        # Get the stash commit
        stash_commit = self._repo.get_object(entry.new_sha)
        assert isinstance(stash_commit, Commit)

        # The stash commit has the working tree changes
        # Its first parent is the commit the stash was based on
        # Its second parent is the index commit
        if len(stash_commit.parents) < 1:
            raise ValueError("Invalid stash entry: no parent commits")

        base_commit_sha = stash_commit.parents[0]

        # Get current HEAD to determine if we can apply cleanly
        try:
            from dulwich.refs import HEADREF

            current_head = self._repo.refs[HEADREF]
        except KeyError:
            raise ValueError("Cannot pop stash: no HEAD")

        # Check if we're at the same commit where the stash was created
        # If not, we need to do a three-way merge
        if current_head != base_commit_sha:
            # For now, we'll apply changes directly but this could cause conflicts
            # A full implementation would do a three-way merge
            pass

        # Apply the stash changes to the working tree and index
        # Get config for working directory update
        config = self._repo.get_config()
        honor_filemode = config.get_boolean(b"core", b"filemode", os.name != "nt")
        validate_path_element = get_path_element_validator(config)

        if config.get_boolean(b"core", b"symlinks", True):
            symlink_fn = symlink
        else:

            def symlink_fn(  # type: ignore[misc,unused-ignore]
                src: str | bytes,
                dst: str | bytes,
                target_is_directory: bool = False,
                *,
                dir_fd: int | None = None,
            ) -> None:
                mode = "w" + ("b" if isinstance(src, bytes) else "")
                with open(dst, mode) as f:
                    f.write(src)

        # Get blob normalizer for line ending conversion
        blob_normalizer = self._repo.get_blob_normalizer(config=config)

        # Open the index
        repo_index = self._repo.open_index(config=config)

        # Apply working tree changes
        stash_tree_id = stash_commit.tree
        repo_path = os.fsencode(self._repo.path)

        # First, if we have index changes (second parent), restore the index state
        if len(stash_commit.parents) >= 2:
            index_commit_sha = stash_commit.parents[1]
            index_commit = self._repo.get_object(index_commit_sha)
            assert isinstance(index_commit, Commit)
            index_tree_id = index_commit.tree

            # Update index entries from the stashed index tree
            tree_entry: TreeEntry
            for tree_entry in iter_tree_contents(
                self._repo.object_store, index_tree_id
            ):
                assert (
                    tree_entry.path is not None
                    and tree_entry.mode is not None
                    and tree_entry.sha is not None
                )
                if not validate_path(tree_entry.path, validate_path_element):
                    raise InvalidPathError(tree_entry.path)

                # Add to index with stage 0 (normal)
                # Get file stats for the entry
                full_path = _tree_to_fs_path(repo_path, tree_entry.path)
                try:
                    st = os.lstat(full_path)
                except FileNotFoundError:
                    # File doesn't exist yet, use dummy stats
                    st = os.stat_result((tree_entry.mode, 0, 0, 0, 0, 0, 0, 0, 0, 0))
                repo_index[tree_entry.path] = index_entry_from_stat(st, tree_entry.sha)

        # Apply working tree changes from the stash
        tree_entry2: TreeEntry
        for tree_entry2 in iter_tree_contents(self._repo.object_store, stash_tree_id):
            assert (
                tree_entry2.path is not None
                and tree_entry2.mode is not None
                and tree_entry2.sha is not None
            )
            if not validate_path(tree_entry2.path, validate_path_element):
                raise InvalidPathError(tree_entry2.path)

            full_path = _tree_to_fs_path(repo_path, tree_entry2.path)

            # Create parent directories if needed
            parent_dir = os.path.dirname(full_path)
            if parent_dir and not os.path.exists(parent_dir):
                os.makedirs(parent_dir)

            # Write the file
            if tree_entry2.mode == S_IFGITLINK:
                # Submodule - just create directory
                if not os.path.isdir(full_path):
                    os.mkdir(full_path)
                st = os.lstat(full_path)
            else:
                obj = self._repo.object_store[tree_entry2.sha]
                assert isinstance(obj, Blob)
                # Apply blob normalization for checkout if normalizer is provided
                if blob_normalizer is not None:
                    obj = blob_normalizer.checkout_normalize(obj, tree_entry2.path)
                st = build_file_from_blob(
                    obj,
                    tree_entry2.mode,
                    full_path,
                    honor_filemode=honor_filemode,
                    symlink_fn=symlink_fn,  # type: ignore[arg-type,unused-ignore]
                )

            # Update index if the file wasn't already staged
            if tree_entry2.path not in repo_index:
                # Update with file stats from disk
                repo_index[tree_entry2.path] = index_entry_from_stat(
                    st, tree_entry2.sha
                )
            else:
                existing_entry = repo_index[tree_entry2.path]

                if (
                    isinstance(existing_entry, IndexEntry)
                    and existing_entry.mode == tree_entry2.mode
                    and existing_entry.sha == tree_entry2.sha
                ):
                    # Update with file stats from disk
                    repo_index[tree_entry2.path] = index_entry_from_stat(
                        st, tree_entry2.sha
                    )

        # Write the updated index
        repo_index.write()

        # Remove the stash entry
        self.drop(index)

        return entry

    def push(
        self,
        committer: bytes | None = None,
        author: bytes | None = None,
        message: bytes | None = None,
        *,
        config: "Config | None" = None,
    ) -> ObjectID:
        """Create a new stash.

        Args:
          committer: Optional committer name to use
          author: Optional author name to use
          message: Optional commit message
          config: Repository configuration. If None, falls back to
            ``self._repo.get_config_stack()``.
        """
        if config is None:
            config = self._repo.get_config_stack()
        # First, create the index commit.
        commit_kwargs = CommitKwargs()
        if committer is not None:
            commit_kwargs["committer"] = committer
        if author is not None:
            commit_kwargs["author"] = author

        index = self._repo.open_index(config=config)
        index_tree_id = index.commit(self._repo.object_store)
        # Create a dangling commit for the index state
        # Note: We pass ref=None which is handled specially in do_commit
        # to create a commit without updating any reference
        index_commit_id = self._repo.get_worktree().commit(
            tree=index_tree_id,
            message=b"Index stash",
            merge_heads=[self._repo.head()],
            no_verify=True,
            sign=False,
            ref=None,  # Don't update any ref
            config=config,
            **commit_kwargs,
        )

        # Then, the working tree one.
        # Filter out entries with None values since commit_tree expects non-None values
        fresh_objects = [
            (path, sha, mode)
            for path, sha, mode in iter_fresh_objects(
                index,
                os.fsencode(self._repo.path),
                object_store=self._repo.object_store,
            )
            if sha is not None and mode is not None
        ]
        stash_tree_id = commit_tree(
            self._repo.object_store,
            fresh_objects,
        )

        if message is None:
            message = b"A stash on " + self._repo.head()

        # TODO(jelmer): Just pass parents into do_commit()?
        self._repo.refs[self._ref] = self._repo.head()

        cid: ObjectID = self._repo.get_worktree().commit(
            ref=self._ref,
            tree=stash_tree_id,
            message=message,
            merge_heads=[index_commit_id],
            no_verify=True,
            sign=False,
            config=config,
            **commit_kwargs,
        )

        # Reset working tree and index to HEAD to match git's behavior
        # Use update_working_tree to reset from stash tree to HEAD tree
        # Get HEAD tree
        head_commit = self._repo.get_object(self._repo.head())
        assert isinstance(head_commit, Commit)
        head_tree_id = head_commit.tree

        # Update from stash tree to HEAD tree
        # This will remove files that were in stash but not in HEAD,
        # and restore files to their HEAD versions
        changes = tree_changes(self._repo.object_store, stash_tree_id, head_tree_id)
        update_working_tree(
            self._repo,
            old_tree_id=stash_tree_id,
            new_tree_id=head_tree_id,
            change_iterator=changes,
            allow_overwrite_modified=True,  # We need to overwrite modified files
            config=config,
        )

        return cid

    def __getitem__(self, index: int) -> "Entry":
        """Get stash entry by index."""
        return list(self.stashes())[index]

    def __len__(self) -> int:
        """Return number of stash entries."""
        return len(list(self.stashes()))



# test_midx.py -- Tests for multi-pack-index
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for multi-pack-index (MIDX) functionality."""

import os
import tempfile
from io import BytesIO
from unittest import TestCase

from dulwich.midx import (
    HASH_ALGORITHM_SHA1,
    MultiPackIndex,
    write_midx,
    write_midx_file,
)


class MIDXWriteTests(TestCase):
    """Tests for writing MIDX files."""

    def test_write_empty_midx(self):
        """Test writing an empty MIDX file."""
        f = BytesIO()
        pack_entries = []
        checksum = write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)

        # Checksum should be 20 bytes
        self.assertEqual(20, len(checksum))

        # Should be able to read it back
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())
        self.assertEqual(0, len(midx))
        self.assertEqual(0, midx.pack_count)
        self.assertEqual([], midx.pack_names)

    def test_write_single_pack_midx(self):
        """Test writing a MIDX file with a single pack."""
        f = BytesIO()

        # Create some fake pack entries
        pack_entries = [
            (
                "pack-abc123.idx",
                [
                    (b"\x01" * 20, 100, 0x12345678),  # sha, offset, crc32
                    (b"\x02" * 20, 200, 0x87654321),
                    (b"\x03" * 20, 300, 0xABCDEF00),
                ],
            )
        ]

        checksum = write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        self.assertEqual(20, len(checksum))

        # Read it back
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        self.assertEqual(3, len(midx))
        self.assertEqual(1, midx.pack_count)
        self.assertEqual(["pack-abc123.idx"], midx.pack_names)

        # Check object lookups
        result = midx.object_offset(b"\x01" * 20)
        self.assertIsNotNone(result)
        pack_name, offset = result
        self.assertEqual("pack-abc123.idx", pack_name)
        self.assertEqual(100, offset)

        result = midx.object_offset(b"\x02" * 20)
        self.assertIsNotNone(result)
        pack_name, offset = result
        self.assertEqual("pack-abc123.idx", pack_name)
        self.assertEqual(200, offset)

        result = midx.object_offset(b"\x03" * 20)
        self.assertIsNotNone(result)
        pack_name, offset = result
        self.assertEqual("pack-abc123.idx", pack_name)
        self.assertEqual(300, offset)

        # Check non-existent object
        result = midx.object_offset(b"\xff" * 20)
        self.assertIsNone(result)

    def test_write_multiple_packs_midx(self):
        """Test writing a MIDX file with multiple packs."""
        f = BytesIO()

        pack_entries = [
            (
                "pack-111.idx",
                [
                    (b"\x01" * 20, 100, 0),
                    (b"\x03" * 20, 300, 0),
                ],
            ),
            (
                "pack-222.idx",
                [
                    (b"\x02" * 20, 50, 0),
                    (b"\x04" * 20, 150, 0),
                ],
            ),
        ]

        checksum = write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        self.assertEqual(20, len(checksum))

        # Read it back
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        self.assertEqual(4, len(midx))
        self.assertEqual(2, midx.pack_count)
        self.assertEqual(["pack-111.idx", "pack-222.idx"], midx.pack_names)

        # Objects should be findable across packs
        result = midx.object_offset(b"\x01" * 20)
        self.assertIsNotNone(result)
        self.assertEqual("pack-111.idx", result[0])

        result = midx.object_offset(b"\x02" * 20)
        self.assertIsNotNone(result)
        self.assertEqual("pack-222.idx", result[0])

    def test_write_large_offsets(self):
        """Test writing a MIDX file with large offsets (>= 2^31)."""
        f = BytesIO()

        large_offset = 2**32  # Offset that requires LOFF chunk
        pack_entries = [
            (
                "pack-large.idx",
                [
                    (b"\x01" * 20, 100, 0),
                    (b"\x02" * 20, large_offset, 0),  # Large offset
                ],
            )
        ]

        checksum = write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        self.assertEqual(20, len(checksum))

        # Read it back
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        self.assertEqual(2, len(midx))

        # Small offset should work
        result = midx.object_offset(b"\x01" * 20)
        self.assertIsNotNone(result)
        self.assertEqual(100, result[1])

        # Large offset should work
        result = midx.object_offset(b"\x02" * 20)
        self.assertIsNotNone(result)
        self.assertEqual(large_offset, result[1])

    def test_write_duplicate_objects_across_packs(self):
        """Objects present in multiple packs must only appear once in OIDL.

        Git's MIDX OIDL chunk requires strictly increasing OIDs; duplicates
        cause 'oid lookup out of order' errors in ``git multi-pack-index
        verify``. This happens in real repositories when cruft packs preserve
        unreachable objects that also exist in other packs.
        """
        f = BytesIO()

        shared_sha = b"\x05" * 20
        pack_entries = [
            (
                "pack-aaa.idx",
                [
                    (b"\x01" * 20, 100, 0),
                    (shared_sha, 200, 0),
                ],
            ),
            (
                "pack-bbb.idx",
                [
                    (shared_sha, 999, 0),
                    (b"\x09" * 20, 50, 0),
                ],
            ),
        ]

        write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        self.assertEqual(3, len(midx))

        # Shared object should resolve to the first pack (pack-aaa sorts first)
        result = midx.object_offset(shared_sha)
        self.assertIsNotNone(result)
        self.assertEqual("pack-aaa.idx", result[0])
        self.assertEqual(200, result[1])

        # OIDs must be strictly increasing (no duplicates)
        entries = list(midx.iterentries())
        shas = [e[0] for e in entries]
        self.assertEqual(shas, sorted(set(shas)))

    def test_write_midx_file(self):
        """Test writing a MIDX file to disk."""
        with tempfile.TemporaryDirectory() as tmpdir:
            midx_path = os.path.join(tmpdir, "multi-pack-index")

            pack_entries = [
                (
                    "pack-test.idx",
                    [
                        (b"\xaa" * 20, 1000, 0),
                    ],
                )
            ]

            checksum = write_midx_file(midx_path, pack_entries, HASH_ALGORITHM_SHA1)
            self.assertEqual(20, len(checksum))

            # Verify file was created
            self.assertTrue(os.path.exists(midx_path))

            # Read it back from disk
            with open(midx_path, "rb") as f:
                midx = MultiPackIndex(midx_path, file=f, contents=f.read())

            self.assertEqual(1, len(midx))
            result = midx.object_offset(b"\xaa" * 20)
            self.assertIsNotNone(result)
            self.assertEqual("pack-test.idx", result[0])
            self.assertEqual(1000, result[1])


class MIDXContainsTests(TestCase):
    """Tests for MIDX __contains__ method."""

    def test_contains_object(self):
        """Test checking if an object is in the MIDX."""
        f = BytesIO()
        pack_entries = [
            (
                "pack-test.idx",
                [
                    (b"\x01" * 20, 100, 0),
                    (b"\x02" * 20, 200, 0),
                ],
            )
        ]

        write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        self.assertTrue(b"\x01" * 20 in midx)
        self.assertTrue(b"\x02" * 20 in midx)
        self.assertFalse(b"\xff" * 20 in midx)

    def test_object_offset_rejects_hex(self):
        """MIDX has a binary-only contract; hex SHAs are not accepted."""
        f = BytesIO()
        pack_entries = [
            (
                "pack-test.idx",
                [
                    (b"\x01" * 20, 100, 0),
                ],
            )
        ]

        write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        # 40-byte hex sha is the wrong length for the binary interface
        self.assertRaises(ValueError, midx.object_offset, b"01" * 20)
        self.assertRaises(ValueError, midx.object_offset, b"\x01" * 10)


class MIDXIterEntriesTests(TestCase):
    """Tests for MIDX iterentries method."""

    def test_iterentries(self):
        """Test iterating over MIDX entries."""
        f = BytesIO()
        pack_entries = [
            (
                "pack-111.idx",
                [
                    (b"\x01" * 20, 100, 0),
                    (b"\x03" * 20, 300, 0),
                ],
            ),
            (
                "pack-222.idx",
                [
                    (b"\x02" * 20, 50, 0),
                ],
            ),
        ]

        write_midx(f, pack_entries, HASH_ALGORITHM_SHA1)
        f.seek(0)
        midx = MultiPackIndex("test.midx", file=f, contents=f.read())

        entries = list(midx.iterentries())
        self.assertEqual(3, len(entries))

        # Entries should be sorted by SHA
        self.assertEqual(b"\x01" * 20, entries[0][0])
        self.assertEqual("pack-111.idx", entries[0][1])
        self.assertEqual(100, entries[0][2])

        self.assertEqual(b"\x02" * 20, entries[1][0])
        self.assertEqual("pack-222.idx", entries[1][1])
        self.assertEqual(50, entries[1][2])

        self.assertEqual(b"\x03" * 20, entries[2][0])
        self.assertEqual("pack-111.idx", entries[2][1])
        self.assertEqual(300, entries[2][2])



# annotate.py -- Annotate files with last changed revision
# Copyright (C) 2015 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Annotate file contents indicating when they were last changed.

Annotated lines are represented as tuples with last modified revision SHA1
and contents.

Please note that this is a very naive annotate implementation. It works,
but its speed could be improved - in particular because it uses
Python's difflib.
"""

__all__ = ["annotate_lines", "update_lines"]

import difflib
from collections.abc import Sequence
from typing import TYPE_CHECKING

from dulwich.objects import Blob
from dulwich.walk import (
    ORDER_DATE,
    Walker,
)

if TYPE_CHECKING:
    from dulwich.diff_tree import TreeChange
    from dulwich.object_store import BaseObjectStore
    from dulwich.objects import Commit, ObjectID, TreeEntry

# Walk over ancestry graph breadth-first
# When checking each revision, find lines that according to difflib.Differ()
# are common between versions.
# Any lines that are not in common were introduced by the newer revision.
# If there were no lines kept from the older version, stop going deeper in the
# graph.


def update_lines(
    annotated_lines: Sequence[tuple[tuple["Commit", "TreeEntry"], bytes]],
    new_history_data: tuple["Commit", "TreeEntry"],
    new_blob: "Blob",
) -> list[tuple[tuple["Commit", "TreeEntry"], bytes]]:
    """Update annotation lines with old blob lines."""
    ret: list[tuple[tuple[Commit, TreeEntry], bytes]] = []
    new_lines = new_blob.splitlines()
    matcher = difflib.SequenceMatcher(
        a=[line for (h, line) in annotated_lines], b=new_lines
    )
    for tag, i1, i2, j1, j2 in matcher.get_opcodes():
        if tag == "equal":
            ret.extend(annotated_lines[i1:i2])
        elif tag in ("insert", "replace"):
            ret.extend([(new_history_data, line) for line in new_lines[j1:j2]])
        elif tag == "delete":
            pass  # don't care
        else:
            raise RuntimeError(f"Unknown tag {tag} returned in diff")
    return ret


def annotate_lines(
    store: "BaseObjectStore",
    commit_id: "ObjectID",
    path: bytes,
    order: str = ORDER_DATE,
    lines: Sequence[tuple[tuple["Commit", "TreeEntry"], bytes]] | None = None,
    follow: bool = True,
) -> list[tuple[tuple["Commit", "TreeEntry"], bytes]]:
    """Annotate the lines of a blob.

    :param store: Object store to retrieve objects from
    :param commit_id: Commit id in which to annotate path
    :param path: Path to annotate
    :param order: Order in which to process history (defaults to ORDER_DATE)
    :param lines: Initial lines to compare to (defaults to specified)
    :param follow: Whether to follow changes across renames/copies
    :return: List of (commit, line) entries where
        commit is the oldest commit that changed a line
    """
    walker = Walker(
        store, include=[commit_id], paths=[path], order=order, follow=follow
    )
    revs: list[tuple[Commit, TreeEntry]] = []
    for log_entry in walker:
        for tree_change in log_entry.changes():
            changes: list[TreeChange]
            if isinstance(tree_change, list):
                changes = tree_change
            else:
                changes = [tree_change]
            for change in changes:
                if change.new is not None and change.new.path == path:
                    if change.old is not None and change.old.path is not None:
                        path = change.old.path
                    revs.append((log_entry.commit, change.new))
                    break

    lines_annotated: list[tuple[tuple[Commit, TreeEntry], bytes]] = []
    for commit, entry in reversed(revs):
        assert entry.sha is not None
        blob_obj = store[entry.sha]
        assert isinstance(blob_obj, Blob)
        lines_annotated = update_lines(lines_annotated, (commit, entry), blob_obj)
    return lines_annotated
